# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common


class ViewBuilder(common.ViewBuilder):
    """Model snapshot API responses as a python dictionary."""

    _collection_name = "snapshots"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, snapshots, snapshot_count=None):
        """Show a list of snapshots without many details."""
        return self._list_view(self.summary, request, snapshots,
                               snapshot_count)

    def detail_list(self, request, snapshots, snapshot_count=None):
        """Detailed view of a list of snapshots."""
        return self._list_view(self.detail, request, snapshots, snapshot_count,
                               coll_name=self._collection_name + '/detail')

    def summary(self, request, snapshot):
        """Generic, non-detailed view of a snapshot."""
        if isinstance(snapshot.metadata, dict):
            metadata = snapshot.metadata
        else:
            metadata = {}

        return {
            'snapshot': {
                'id': snapshot.id,
                'created_at': snapshot.created_at,
                'updated_at': snapshot.updated_at,
                'name': snapshot.display_name,
                'description': snapshot.display_description,
                'volume_id': snapshot.volume_id,
                'status': snapshot.status,
                'size': snapshot.volume_size,
                'metadata': metadata,
            }
        }

    def detail(self, request, snapshot):
        """Detailed view of a single snapshot."""
        # NOTE(geguileo): No additional data at the moment
        return self.summary(request, snapshot)

    def _list_view(self, func, request, snapshots, snapshot_count,
                   coll_name=_collection_name):
        """Provide a view for a list of snapshots."""
        snapshots_list = [func(request, snapshot)['snapshot']
                          for snapshot in snapshots]
        snapshots_links = self._get_collection_links(request,
                                                     snapshots,
                                                     coll_name,
                                                     snapshot_count)
        snapshots_dict = {self._collection_name: snapshots_list}

        if snapshots_links:
            snapshots_dict[self._collection_name + '_links'] = snapshots_links

        if snapshot_count is not None:
            snapshots_dict['count'] = snapshot_count

        return snapshots_dict



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for V3 scheduler_hints API.

"""

from cinder.api.validation import parameter_types

create = {
    'type': 'object',
    'properties': {
        'OS-SCH-HNT:scheduler_hints': {
            'type': ['object', 'null'],
            'properties': {
                'local_to_instance': parameter_types.optional_uuid,
                'different_host': {
                    # NOTE: The value of 'different_host' is the set of volume
                    # uuids where a new volume is scheduled on a different
                    # host. A user can specify one volume as string parameter
                    # and should specify multiple volumes as array parameter
                    # instead.
                    'oneOf': [
                        {
                            'type': 'string',
                            'format': 'uuid'
                        },
                        {
                            'type': 'array',
                            'items': parameter_types.uuid,
                            'uniqueItems': True,
                        }
                    ]
                },
                'same_host': {
                    # NOTE: The value of 'same_host' is the set of volume
                    # uuids where a new volume is scheduled on the same host.
                    # A user can specify one volume as string parameter and
                    # should specify multiple volumes as array parameter
                    # instead.
                    'oneOf': [
                        {
                            'type': 'string',
                            'format': 'uuid'
                        },
                        {
                            'type': 'array',
                            'items': parameter_types.uuid,
                            'uniqueItems': True,
                        }
                    ]
                },
                'query': {
                    # NOTE: The value of 'query' is converted to dict data with
                    # jsonutils.loads() and used for filtering hosts.
                    'type': ['string', 'object'],
                },
            },
            # NOTE: As this Mail:
            # http://lists.openstack.org/pipermail/openstack-dev/2015-June/067996.html
            # pointed out the limit the scheduler-hints in the API is
            # problematic. So relax it.
            'additionalProperties': True
        },
    },
}



# Copyright (c) 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinder.api.contrib import types_extra_specs
from cinder.api import microversions as mv
from cinder.api.v3 import types
from cinder.policies import type_extra_specs as policy
from cinder.policies import volume_type as type_policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils


@ddt.ddt
class TypeExtraSpecsPolicyTest(base.BasePolicyTest):
    """Verify extra specs policy settings for the types API"""

    # Deprecated check_str="" allows anyone to read extra specs
    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_readers = [
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_exceptions = []

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = types_extra_specs.VolumeTypeExtraSpecsController()
        self.api_path = '/v3/%s/types' % (self.project_id)
        self.api_version = mv.BASE_VERSION

    @ddt.data(*base.all_users)
    def test_get_all_policy(self, user_id):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type',
                                                 extra_specs={'foo': 'bar'})
        rule_name = policy.GET_ALL_POLICY
        url = '%s/%s/extra_specs' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        self.common_policy_check(user_id,
                                 self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller.index,
                                 req,
                                 type_id=vol_type.id)

    @ddt.data(*base.all_users)
    def test_get_policy(self, user_id):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type',
                                                 extra_specs={'foo': 'bar'})
        rule_name = policy.GET_POLICY
        url = '%s/%s/extra_specs/foo' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        # Relax the READ_SENSITIVE_POLICY policy so that any user is able
        # to "see" the spec.
        self.policy.set_rules({policy.READ_SENSITIVE_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id,
                                 self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller.show,
                                 req,
                                 type_id=vol_type.id,
                                 id='foo')

    @ddt.data(*base.all_users)
    def test_create_policy(self, user_id):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type')
        rule_name = policy.CREATE_POLICY
        url = '%s/%s/extra_specs' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "extra_specs": {
                "foo": "bar",
            }
        }

        self.common_policy_check(user_id,
                                 self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller.create,
                                 req,
                                 type_id=vol_type.id,
                                 body=body)

    @ddt.data(*base.all_users)
    def test_update_policy(self, user_id):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type',
                                                 extra_specs={'foo': 'bar'})
        rule_name = policy.UPDATE_POLICY
        url = '%s/%s/extra_specs/foo' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'PUT'
        body = {"foo": "zap"}

        self.common_policy_check(user_id,
                                 self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller.update,
                                 req,
                                 type_id=vol_type.id,
                                 id='foo',
                                 body=body)

    @ddt.data(*base.all_users)
    def test_delete_policy(self, user_id):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type',
                                                 extra_specs={'foo': 'bar'})
        rule_name = policy.DELETE_POLICY
        url = '%s/%s/extra_specs/foo' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'

        self.common_policy_check(user_id,
                                 self.authorized_admins,
                                 self.unauthorized_admins,
                                 self.unauthorized_exceptions,
                                 rule_name,
                                 self.controller.delete,
                                 req,
                                 type_id=vol_type.id,
                                 id='foo')

    @ddt.data(*base.all_users)
    def test_read_sensitive_policy(self, user_id):
        # The 'multiattach' extra spec is user visible, and the
        # 'sensitive' extra spec should not be user visible.
        extra_specs = {
            'multiattach': '<is> True',
            'sensitive': 'secret',
        }
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 testcase_instance=self,
                                                 name='fake_vol_type',
                                                 extra_specs=extra_specs)
        rule_name = policy.READ_SENSITIVE_POLICY
        url = '%s/%s' % (self.api_path, vol_type.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        # Relax these policies in order to get past those checks.
        self.policy.set_rules({type_policy.GET_POLICY: ""},
                              overwrite=False)
        self.policy.set_rules({type_policy.EXTRA_SPEC_POLICY: ""},
                              overwrite=False)

        # With the relaxed policies, all users are authorized because
        # failing the READ_SENSITIVE_POLICY policy check is not fatal.
        authorized_users = [user_id]
        unauthorized_users = []

        controller = types.VolumeTypesController()
        response = self.common_policy_check(user_id,
                                            authorized_users,
                                            unauthorized_users,
                                            self.unauthorized_exceptions,
                                            rule_name,
                                            controller.show,
                                            req,
                                            id=vol_type.id)

        if user_id in self.authorized_admins:
            # Admins should see all extra specs
            expected = extra_specs
        else:
            # Non-admins should only see user visible extra specs
            expected = {'multiattach': '<is> True'}
        self.assertDictEqual(expected, response['volume_type']['extra_specs'])


class TypeExtraSpecsPolicySecureRbacTest(TypeExtraSpecsPolicyTest):
    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'system_member',
        'system_reader',
        'other_project_member',
        'other_project_reader',
    ]

    unauthorized_readers = [
        # These are unauthorized because they don't have the reader role
        'legacy_owner',
        'project_foo',
        'system_foo',
    ]

    # NOTE(Xena): The authorized_admins and unauthorized_admins are the same
    # as the TypeExtraSpecsPolicyTest. This is because in Xena the "admin only"
    # rules are the legacy RULE_ADMIN_API. This will change in Yoga, when
    # RULE_ADMIN_API will be deprecated in favor of the SYSTEM_ADMIN rule that
    # is scope based.
    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



# Copyright (c) 2012 NetApp, Inc.  All rights reserved.
# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# Copyright (c) 2016 Michael Price.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utilities for NetApp drivers.

This module contains common utilities to be used by one or more
NetApp drivers to achieve the desired functionality.
"""


import decimal
import platform
import re

from oslo_concurrency import processutils as putils
from oslo_log import log as logging
from oslo_utils import netutils

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder import version
from cinder.volume import qos_specs
from cinder.volume import volume_types


LOG = logging.getLogger(__name__)


OPENSTACK_PREFIX = 'openstack-'
OBSOLETE_SSC_SPECS = {'netapp:raid_type': 'netapp_raid_type',
                      'netapp:disk_type': 'netapp_disk_type'}
DEPRECATED_SSC_SPECS = {'netapp_unmirrored': 'netapp_mirrored',
                        'netapp_nodedup': 'netapp_dedup',
                        'netapp_nocompression': 'netapp_compression',
                        'netapp_thick_provisioned': 'netapp_thin_provisioned'}
MIN_QOS_KEYS = frozenset([
    'minIOPS',
    'minIOPSperGiB',
])
MAX_QOS_KEYS = frozenset([
    'maxIOPS',
    'maxIOPSperGiB',
    'maxBPS',
    'maxBPSperGiB',
])
ADAPTIVE_QOS_KEYS = frozenset([
    'expectedIOPSperGiB',
    'peakIOPSperGiB',
    'expectedIOPSAllocation',
    'peakIOPSAllocation',
    'absoluteMinIOPS',
    'blockSize',
])
QOS_ADAPTIVE_POLICY_GROUP_SPEC_KEYS = frozenset([
    'expected_iops',
    'peak_iops',
    'expected_iops_allocation',
    'peak_iops_allocation',
    'absolute_min_iops',
    'block_size',
    'policy_name',
])
BACKEND_QOS_CONSUMERS = frozenset(['back-end', 'both'])

# Secret length cannot be less than 96 bits. http://tools.ietf.org/html/rfc3723
CHAP_SECRET_LENGTH = 16
DEFAULT_CHAP_USER_NAME = 'NetApp_iSCSI_CHAP_Username'
API_TRACE_PATTERN = '(.*)'


class NetAppDriverException(exception.VolumeDriverException):
    message = _("NetApp Cinder Driver exception.")


class GeometryHasChangedOnDestination(NetAppDriverException):
    message = _("Geometry has changed on destination volume.")


class NetAppDriverTimeout(NetAppDriverException):
    message = _("Timeout in NetApp Cinder Driver.")


def validate_instantiation(**kwargs):
    """Checks if a driver is instantiated other than by the unified driver.

    Helps check direct instantiation of netapp drivers.
    Call this function in every netapp block driver constructor.
    """
    if kwargs and kwargs.get('netapp_mode') == 'proxy':
        return
    LOG.warning("It is not the recommended way to use drivers by NetApp. "
                "Please use NetAppDriver to achieve the functionality.")


def check_flags(required_flags, configuration):
    """Ensure that the flags we care about are set."""
    for flag in required_flags:
        if not getattr(configuration, flag, None):
            msg = _('Configuration value %s is not set.') % flag
            raise exception.InvalidInput(reason=msg)


def to_bool(val):
    """Converts true, yes, y, 1 to True, False otherwise."""
    if val:
        strg = str(val).lower()
        if (strg == 'true' or strg == 'y'
            or strg == 'yes' or strg == 'enabled'
                or strg == '1'):
            return True
        else:
            return False
    else:
        return False


@utils.synchronized("safe_set_attr")
def set_safe_attr(instance, attr, val):
    """Sets the attribute in a thread safe manner.

    Returns if new val was set on attribute.
    If attr already had the value then False.
    """

    if not instance or not attr:
        return False
    old_val = getattr(instance, attr, None)
    if val is None and old_val is None:
        return False
    elif val == old_val:
        return False
    else:
        setattr(instance, attr, val)
        return True


def get_volume_extra_specs(volume):
    """Provides extra specs associated with volume."""
    ctxt = context.get_admin_context()
    type_id = volume.get('volume_type_id')
    if type_id is None:
        return {}
    volume_type = volume_types.get_volume_type(ctxt, type_id)
    if volume_type is None:
        return {}
    extra_specs = volume_type.get('extra_specs', {})
    log_extra_spec_warnings(extra_specs)
    return extra_specs


def setup_api_trace_pattern(api_trace_pattern):
    global API_TRACE_PATTERN
    try:
        re.compile(api_trace_pattern)
    except (re.error, TypeError):
        msg = _('Cannot parse the API trace pattern. %s is not a '
                'valid python regular expression.') % api_trace_pattern
        raise exception.InvalidConfigurationValue(msg)
    API_TRACE_PATTERN = api_trace_pattern


def trace_filter_func_api(all_args):
    na_element = all_args.get('na_element')
    if na_element is None:
        return True
    api_name = na_element.get_name()
    return re.match(API_TRACE_PATTERN, api_name) is not None


def trace_filter_func_rest_api(all_args):
    url = all_args.get('url')
    if url is None:
        return True
    return re.match(API_TRACE_PATTERN, url) is not None


def round_down(value, precision='0.00'):
    return float(decimal.Decimal(str(value)).quantize(
        decimal.Decimal(precision), rounding=decimal.ROUND_DOWN))


def log_extra_spec_warnings(extra_specs):
    for spec in (set(extra_specs.keys() if extra_specs else []) &
                 set(OBSOLETE_SSC_SPECS.keys())):
        LOG.warning('Extra spec %(old)s is obsolete.  Use %(new)s '
                    'instead.', {'old': spec,
                                 'new': OBSOLETE_SSC_SPECS[spec]})
    for spec in (set(extra_specs.keys() if extra_specs else []) &
                 set(DEPRECATED_SSC_SPECS.keys())):
        LOG.warning('Extra spec %(old)s is deprecated.  Use %(new)s '
                    'instead.', {'old': spec,
                                 'new': DEPRECATED_SSC_SPECS[spec]})


def get_iscsi_connection_properties(lun_id, volume, iqns,
                                    addresses, ports):
    # literal ipv6 address
    addresses = [netutils.escape_ipv6(a) if netutils.is_valid_ipv6(a) else a
                 for a in addresses]

    lun_id = int(lun_id)
    if isinstance(iqns, str):
        iqns = [iqns] * len(addresses)

    target_portals = ['%s:%s' % (a, p) for a, p in zip(addresses, ports)]

    properties = {}
    properties['target_discovered'] = False
    properties['target_portal'] = target_portals[0]
    properties['target_iqn'] = iqns[0]
    properties['target_lun'] = lun_id
    properties['volume_id'] = volume['id']
    if len(addresses) > 1:
        properties['target_portals'] = target_portals
        properties['target_iqns'] = iqns
        properties['target_luns'] = [lun_id] * len(addresses)

    auth = volume['provider_auth']
    if auth:
        (auth_method, auth_username, auth_secret) = auth.split()
        properties['auth_method'] = auth_method
        properties['auth_username'] = auth_username
        properties['auth_password'] = auth_secret
    return {
        'driver_volume_type': 'iscsi',
        'data': properties,
    }


def validate_qos_spec(qos_spec):
    """Check validity of Cinder qos spec for our backend."""
    if qos_spec is None:
        return

    normalized_min_keys = [key.lower() for key in MIN_QOS_KEYS]
    normalized_max_keys = [key.lower() for key in MAX_QOS_KEYS]
    normalized_aqos_keys = [key.lower() for key in ADAPTIVE_QOS_KEYS]

    unrecognized_keys = [
        k for k in qos_spec.keys()
        if k.lower() not in
        normalized_max_keys + normalized_min_keys + normalized_aqos_keys]

    if unrecognized_keys:
        msg = _('Unrecognized QOS keywords: "%s"') % unrecognized_keys
        raise exception.Invalid(msg)

    min_dict = {k: v for k, v in qos_spec.items()
                if k.lower() in normalized_min_keys}
    if len(min_dict) > 1:
        msg = _('Only one minimum limit can be set in a QoS spec.')
        raise exception.Invalid(msg)

    max_dict = {k: v for k, v in qos_spec.items()
                if k.lower() in normalized_max_keys}
    if len(max_dict) > 1:
        msg = _('Only one maximum limit can be set in a QoS spec.')
        raise exception.Invalid(msg)

    aqos_dict = {k: v for k, v in qos_spec.items()
                 if k.lower() in normalized_aqos_keys}
    if aqos_dict and (min_dict or max_dict):
        msg = _('Adaptive QoS specs and non-adaptive QoS specs '
                'cannot be used together.')
        raise exception.Invalid(msg)


def get_volume_type_from_volume(volume):
    """Provides volume type associated with volume."""
    type_id = volume.get('volume_type_id')
    if type_id is None:
        return {}
    ctxt = context.get_admin_context()
    return volume_types.get_volume_type(ctxt, type_id)


def _get_min_throughput_from_qos_spec(qos_spec, volume_size):
    """Returns the minimum QoS throughput.

    The QoS min specs are exclusive of one another and it accepts values in
    IOPS only.
    """
    if 'miniops' in qos_spec:
        min_throughput = '%siops' % qos_spec['miniops']
    elif 'miniopspergib' in qos_spec:
        min_throughput = '%siops' % str(
            int(qos_spec['miniopspergib']) * int(volume_size))
    else:
        min_throughput = None
    return min_throughput


def _get_max_throughput_from_qos_spec(qos_spec, volume_size):
    """Returns the maximum QoS throughput.

    The QoS max specs are exclusive of one another.
    """
    if 'maxiops' in qos_spec:
        max_throughput = '%siops' % qos_spec['maxiops']
    elif 'maxiopspergib' in qos_spec:
        max_throughput = '%siops' % str(
            int(qos_spec['maxiopspergib']) * int(volume_size))
    elif 'maxbps' in qos_spec:
        max_throughput = '%sB/s' % qos_spec['maxbps']
    elif 'maxbpspergib' in qos_spec:
        max_throughput = '%sB/s' % str(
            int(qos_spec['maxbpspergib']) * int(volume_size))
    else:
        max_throughput = None
    return max_throughput


def map_qos_spec(qos_spec, volume):
    """Map Cinder QOS spec to limit/throughput-value as used in client API."""
    if qos_spec is None:
        return None

    spec = map_dict_to_lower(qos_spec)
    min_throughput = _get_min_throughput_from_qos_spec(spec, volume['size'])
    max_throughput = _get_max_throughput_from_qos_spec(spec, volume['size'])

    if min_throughput and max_throughput and max_throughput.endswith('B/s'):
        msg = _('Maximum limit should be in IOPS when minimum limit is '
                'specified.')
        raise exception.Invalid(msg)

    if min_throughput and max_throughput and max_throughput < min_throughput:
        msg = _('Maximum limit should be greater than or equal to the '
                'minimum limit.')
        raise exception.Invalid(msg)

    policy = dict(policy_name=get_qos_policy_group_name(volume))
    if min_throughput:
        policy['min_throughput'] = min_throughput
    if max_throughput:
        policy['max_throughput'] = max_throughput
    return policy


def map_aqos_spec(qos_spec, volume):
    """Map Cinder QOS spec to Adaptive QoS values."""
    if qos_spec is None:
        return None

    qos_spec = map_dict_to_lower(qos_spec)
    spec = dict(policy_name=get_qos_policy_group_name(volume))

    # Adaptive QoS specs
    if 'expectediopspergib' in qos_spec:
        spec['expected_iops'] = (
            '%sIOPS/GB' % qos_spec['expectediopspergib'])
    if 'peakiopspergib' in qos_spec:
        spec['peak_iops'] = '%sIOPS/GB' % qos_spec['peakiopspergib']
    if 'expectediopsallocation' in qos_spec:
        spec['expected_iops_allocation'] = qos_spec['expectediopsallocation']
    if 'peakiopsallocation' in qos_spec:
        spec['peak_iops_allocation'] = qos_spec['peakiopsallocation']
    if 'absoluteminiops' in qos_spec:
        spec['absolute_min_iops'] = '%sIOPS' % qos_spec['absoluteminiops']
    if 'blocksize' in qos_spec:
        spec['block_size'] = qos_spec['blocksize']

    if 'peak_iops' not in spec or 'expected_iops' not in spec:
        msg = _('Adaptive QoS requires the expected property and '
                'the peak property set together.')
        raise exception.Invalid(msg)

    if spec['peak_iops'] < spec['expected_iops']:
        msg = _('Adaptive maximum limit should be greater than or equal to '
                'the adaptive minimum limit.')
        raise exception.Invalid(msg)

    return spec


def map_dict_to_lower(input_dict):
    """Return an equivalent to the input dictionary with lower-case keys."""
    lower_case_dict = {}
    for key in input_dict:
        lower_case_dict[key.lower()] = input_dict[key]
    return lower_case_dict


def get_qos_policy_group_name(volume):
    """Return the name of backend QOS policy group based on its volume id."""
    if 'id' in volume:
        return OPENSTACK_PREFIX + volume.name_id
    return None


def get_qos_policy_group_name_from_info(qos_policy_group_info):
    """Return the name of a QOS policy group given qos policy group info."""
    if qos_policy_group_info is None:
        return None
    legacy = qos_policy_group_info.get('legacy')
    if legacy is not None:
        return legacy['policy_name']
    spec = qos_policy_group_info.get('spec')
    if spec is not None:
        return spec['policy_name']
    return None


def get_pool_name_filter_regex(configuration):
    """Build the regex for filtering pools by name

    :param configuration: The volume driver configuration
    :raise InvalidConfigurationValue: if configured regex pattern is invalid
    :returns: A compiled regex for filtering pool names
    """

    # If the configuration parameter is specified as an empty string
    # (interpreted as matching all pools), we replace it here with
    # (.+) to be explicit with CSV compatibility support implemented below.
    pool_patterns = configuration.netapp_pool_name_search_pattern or r'(.+)'

    # Strip whitespace from start/end and then 'or' all regex patterns
    pool_patterns = '|'.join(['^' + pool_pattern.strip('^$ \t') + '$' for
                              pool_pattern in pool_patterns.split(',')])

    try:
        return re.compile(pool_patterns)
    except re.error:
        raise exception.InvalidConfigurationValue(
            option='netapp_pool_name_search_pattern',
            value=configuration.netapp_pool_name_search_pattern)


def get_valid_qos_policy_group_info(volume, extra_specs=None):
    """Given a volume, return information for QOS provisioning."""
    info = dict(legacy=None, spec=None)
    try:
        volume_type = get_volume_type_from_volume(volume)
    except (KeyError, exception.NotFound):
        LOG.exception('Cannot get QoS spec for volume %s.', volume['id'])
        return info
    if volume_type is None:
        return info
    if extra_specs is None:
        extra_specs = volume_type.get('extra_specs', {})
    info['legacy'] = get_legacy_qos_policy(extra_specs)
    info['spec'] = get_valid_backend_qos_spec_from_volume_type(volume,
                                                               volume_type)
    msg = 'QoS policy group info for volume %(vol)s: %(info)s'
    LOG.debug(msg, {'vol': volume['name'], 'info': info})
    check_for_invalid_qos_spec_combination(info, volume_type)
    return info


def get_valid_backend_qos_spec_from_volume_type(volume, volume_type):
    """Given a volume type, return the associated Cinder QoS spec."""
    spec_dict = get_backend_qos_spec_from_volume_type(volume_type)
    if spec_dict is None:
        return None
    validate_qos_spec(spec_dict)
    map_spec = (map_aqos_spec
                if is_qos_adaptive(spec_dict)
                else map_qos_spec)
    return map_spec(spec_dict, volume)


def is_qos_adaptive(spec_dict):
    if not spec_dict:
        return False

    normalized_aqos_keys = [key.lower() for key in ADAPTIVE_QOS_KEYS]
    return all(key in normalized_aqos_keys
               for key in map_dict_to_lower(spec_dict).keys())


def is_qos_policy_group_spec_adaptive(policy):
    if not policy:
        return False

    spec = policy.get('spec')
    if not spec:
        return False

    return all(key in QOS_ADAPTIVE_POLICY_GROUP_SPEC_KEYS
               for key in map_dict_to_lower(spec).keys())


def get_backend_qos_spec_from_volume_type(volume_type):
    qos_specs_id = volume_type.get('qos_specs_id')
    if qos_specs_id is None:
        return None
    ctxt = context.get_admin_context()
    qos_spec = qos_specs.get_qos_specs(ctxt, qos_specs_id)
    if qos_spec is None:
        return None
    consumer = qos_spec['consumer']
    # Front end QoS specs are handled by libvirt and we ignore them here.
    if consumer not in BACKEND_QOS_CONSUMERS:
        return None
    return qos_spec['specs']


def check_for_invalid_qos_spec_combination(info, volume_type):
    """Invalidate QOS spec if both legacy and non-legacy info is present."""
    if info['legacy'] and info['spec']:
        msg = _('Conflicting QoS specifications in volume type '
                '%s: when QoS spec is associated to volume '
                'type, legacy "netapp:qos_policy_group" is not allowed in '
                'the volume type extra specs.') % volume_type['id']
        raise exception.Invalid(msg)


def get_legacy_qos_policy(extra_specs):
    """Return legacy qos policy information if present in extra specs."""
    external_policy_name = extra_specs.get('netapp:qos_policy_group')
    if external_policy_name is None:
        return None
    return dict(policy_name=external_policy_name)


def get_export_host_junction_path(share):
    if '[' in share and ']' in share:
        try:
            # ipv6
            host = re.search(r'\[(.*)\]', share).group(1)
            junction_path = share.split(':')[-1]
        except AttributeError:
            raise NetAppDriverException(_("Share '%s' is not in a valid "
                                          "format.") % share)
    else:
        # ipv4
        path = share.split(':')
        if len(path) == 2:
            host = path[0]
            junction_path = path[1]
        else:
            raise NetAppDriverException(_("Share '%s' is not in a valid "
                                          "format.") % share)

    return host, junction_path


def qos_min_feature_name(is_nfs, node_name):
    if node_name is None:
        node_name = ''
    if is_nfs:
        return 'QOS_MIN_NFS_' + node_name
    else:
        return 'QOS_MIN_BLOCK_' + node_name


class hashabledict(dict):
    """A hashable dictionary that is comparable (i.e. in unit tests, etc.)"""
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class OpenStackInfo(object):
    """OS/distribution, release, and version.

    NetApp uses these fields as content for EMS log entry.
    """

    PACKAGE_NAME = 'python-cinder'

    def __init__(self):
        self._version = 'unknown version'
        self._release = 'unknown release'
        self._vendor = 'unknown vendor'
        self._platform = 'unknown platform'

    def _update_version_from_version_string(self):
        try:
            self._version = version.version_info.version_string()
        except Exception:
            pass

    def _update_release_from_release_string(self):
        try:
            self._release = version.version_info.release_string()
        except Exception:
            pass

    def _update_platform(self):
        try:
            self._platform = platform.platform()
        except Exception:
            pass

    @staticmethod
    def _get_version_info_version():
        return version.version_info.version

    @staticmethod
    def _get_version_info_release():
        return version.version_info.release

    def _update_info_from_version_info(self):
        try:
            ver = self._get_version_info_version()
            if ver:
                self._version = ver
        except Exception:
            pass
        try:
            rel = self._get_version_info_release()
            if rel:
                self._release = rel
        except Exception:
            pass

    # RDO, RHEL-OSP, Mirantis on Redhat, SUSE
    def _update_info_from_rpm(self):
        LOG.debug('Trying rpm command.')
        try:
            out, err = putils.execute("rpm", "-q", "--queryformat",
                                      "'%{version}\t%{release}\t%{vendor}'",
                                      self.PACKAGE_NAME)
            if not out:
                LOG.info('No rpm info found for %(pkg)s package.', {
                    'pkg': self.PACKAGE_NAME})
                return False
            parts = out.split()
            self._version = parts[0]
            self._release = parts[1]
            self._vendor = ' '.join(parts[2::])
            return True
        except Exception as e:
            LOG.info('Could not run rpm command: %(msg)s.', {'msg': e})
            return False

    # ubuntu, mirantis on ubuntu
    def _update_info_from_dpkg(self):
        LOG.debug('Trying dpkg-query command.')
        try:
            _vendor = None
            out, err = putils.execute("dpkg-query", "-W", "-f='${Version}'",
                                      self.PACKAGE_NAME)
            if not out:
                LOG.info('No dpkg-query info found for %(pkg)s package.',
                         {'pkg': self.PACKAGE_NAME})
                return False
            # debian format: [epoch:]upstream_version[-debian_revision]
            deb_version = out
            # in case epoch or revision is missing, copy entire string
            _release = deb_version
            if ':' in deb_version:
                deb_epoch, upstream_version = deb_version.split(':')
                _release = upstream_version
            if '-' in deb_version:
                deb_revision = deb_version.split('-')[1]
                _vendor = deb_revision
            self._release = _release
            if _vendor:
                self._vendor = _vendor
            return True
        except Exception as e:
            LOG.info('Could not run dpkg-query command: %(msg)s.', {
                'msg': e})
            return False

    def _update_openstack_info(self):
        self._update_version_from_version_string()
        self._update_release_from_release_string()
        self._update_platform()
        # some distributions override with more meaningful information
        self._update_info_from_version_info()
        # see if we have still more targeted info from rpm or apt
        found_package = self._update_info_from_rpm()
        if not found_package:
            self._update_info_from_dpkg()

    def info(self):
        self._update_openstack_info()
        return '%(version)s|%(release)s|%(vendor)s|%(platform)s' % {
            'version': self._version, 'release': self._release,
            'vendor': self._vendor, 'platform': self._platform}


class Features(object):

    def __init__(self):
        self.defined_features = set()

    def add_feature(self, name, supported=True, min_version=None):
        if not isinstance(supported, bool):
            raise TypeError("Feature value must be a bool type.")
        self.defined_features.add(name)
        setattr(self, name, FeatureState(supported, min_version))

    def __getattr__(self, name):
        # NOTE(cknight): Needed to keep pylint happy.
        raise AttributeError


class FeatureState(object):

    def __init__(self, supported=True, minimum_version=None):
        """Represents the current state of enablement for a Feature

        :param supported: True if supported, false otherwise
        :param minimum_version: The minimum version that this feature is
        supported at
        """
        self.supported = supported
        self.minimum_version = minimum_version

    def __nonzero__(self):
        """Allow a FeatureState object to be tested for truth value

        :returns: True if the feature is supported, otherwise False
        """
        return self.supported

    def __bool__(self):
        """py3 Allow a FeatureState object to be tested for truth value

        :returns: True if the feature is supported, otherwise False
        """
        return self.supported


class BitSet(object):
    def __init__(self, value=0):
        self._value = value

    def set(self, bit):
        self._value |= 1 << bit
        return self

    def unset(self, bit):
        self._value &= ~(1 << bit)
        return self

    def is_set(self, bit):
        return self._value & 1 << bit

    def __and__(self, other):
        self._value &= other
        return self

    def __or__(self, other):
        self._value |= other
        return self

    def __invert__(self):
        self._value = ~self._value
        return self

    def __xor__(self, other):
        self._value ^= other
        return self

    def __lshift__(self, other):
        self._value <<= other
        return self

    def __rshift__(self, other):
        self._value >>= other
        return self

    def __int__(self):
        return self._value

    def __str__(self):
        return bin(self._value)

    def __repr__(self):
        return str(self._value)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self._value ==
                other._value) or self._value == int(other)

    def __ne__(self, other):
        return not self.__eq__(other)



# Copyright (c) 2014 Quobyte Inc.
# Copyright (c) 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Quobyte driver module."""

import errno
import os
import shutil
import traceback
from unittest import mock

import ddt
from oslo_concurrency import processutils as putils
from oslo_utils import fileutils
from oslo_utils import imageutils
from oslo_utils import units
import psutil
import six

from cinder import context
from cinder import db
from cinder import exception
from cinder.image import image_utils
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers import quobyte
from cinder.volume.drivers import remotefs


class FakeDb(object):
    msg = "Tests are broken: mock this out."

    def volume_get(self, *a, **kw):
        raise Exception(self.msg)

    def snapshot_get_all_for_volume(self, *a, **kw):
        """Mock this if you want results from it."""
        return []

    def volume_get_all(self, *a, **kw):
        return []


@ddt.ddt
class QuobyteDriverTestCase(test.TestCase):
    """Test case for Quobyte driver."""

    TEST_QUOBYTE_VOLUME = 'quobyte://quobyte-host/openstack-volumes'
    TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL = 'quobyte-host/openstack-volumes'
    TEST_SIZE_IN_GB = 1
    TEST_MNT_HASH = "1331538734b757ed52d0e18c0a7210cd"
    TEST_MNT_POINT_BASE = '/fake-mnt'
    TEST_MNT_POINT = os.path.join(TEST_MNT_POINT_BASE, TEST_MNT_HASH)
    TEST_FILE_NAME = 'test.txt'
    TEST_SHARES_CONFIG_FILE = '/etc/cinder/test-shares.conf'
    TEST_TMP_FILE = '/tmp/tempfile'
    VOLUME_UUID = 'abcdefab-cdef-abcd-efab-cdefabcdefab'
    SNAP_UUID = 'bacadaca-baca-daca-baca-dacadacadaca'
    SNAP_UUID_2 = 'bebedede-bebe-dede-bebe-dedebebedede'
    CACHE_NAME = quobyte.QuobyteDriver.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME

    def _get_fake_snapshot(self, src_volume):
        snapshot = fake_snapshot.fake_snapshot_obj(
            self.context,
            volume_name=src_volume.name,
            display_name='clone-snap-%s' % src_volume.id,
            size=src_volume.size,
            volume_size=src_volume.size,
            volume_id=src_volume.id,
            id=self.SNAP_UUID)
        snapshot.volume = src_volume
        return snapshot

    def setUp(self):
        super(QuobyteDriverTestCase, self).setUp()

        self._configuration = mock.Mock(conf.Configuration)
        self._configuration.append_config_values(mock.ANY)
        self._configuration.quobyte_volume_url = \
            self.TEST_QUOBYTE_VOLUME
        self._configuration.quobyte_client_cfg = None
        self._configuration.quobyte_sparsed_volumes = True
        self._configuration.quobyte_qcow2_volumes = False
        self._configuration.quobyte_mount_point_base = \
            self.TEST_MNT_POINT_BASE
        self._configuration.nas_secure_file_operations = "true"
        self._configuration.nas_secure_file_permissions = "true"
        self._configuration.quobyte_volume_from_snapshot_cache = False
        self._configuration.quobyte_overlay_volumes = False

        self._driver = quobyte.QuobyteDriver(configuration=self._configuration)
        self._driver.shares = {}
        self._driver.set_nas_security_options(is_new_cinder_install=False)
        self._driver.base = self._configuration.quobyte_mount_point_base

        self.context = context.get_admin_context()

    def assertRaisesAndMessageMatches(
            self, excClass, msg, callableObj, *args, **kwargs):
        """Ensure that the specified exception was raised. """

        caught = False
        try:
            callableObj(*args, **kwargs)
        except Exception as exc:
            caught = True
            self.assertIsInstance(exc, excClass,
                                  'Wrong exception caught: %s Stacktrace: %s' %
                                  (exc, traceback.format_exc()))
            self.assertIn(msg, six.text_type(exc))

        if not caught:
            self.fail('Expected raised exception but nothing caught.')

    def get_mock_partitions(self):
        mypart = mock.Mock()
        mypart.device = "quobyte@"
        mypart.mountpoint = self.TEST_MNT_POINT
        return [mypart]

    @mock.patch.object(os, "symlink")
    def test__create_overlay_volume_from_snapshot(self, os_sl_mock):
        drv = self._driver
        drv._execute = mock.Mock()
        vol = self._simple_volume()
        snap = self._get_fake_snapshot(vol)
        r_path = os.path.join(drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                              snap.id)
        vol_path = drv._local_path_volume(vol)

        drv._create_overlay_volume_from_snapshot(vol, snap, 1, "qcow2")

        drv._execute.assert_called_once_with(
            'qemu-img', 'create', '-f', 'qcow2', '-o',
            'backing_file=%s,backing_fmt=qcow2' % (r_path), vol_path, "1G",
            run_as_root=drv._execute_as_root)
        os_sl_mock.assert_called_once_with(
            drv.local_path(vol),
            drv._local_volume_from_snap_cache_path(snap) + '.child-' + vol.id)

    def test__create_regular_file(self):
        with mock.patch.object(self._driver, "_execute") as qb_exec_mock:
            tmp_path = "/path/for/test"
            test_size = 1

            self._driver._create_regular_file(tmp_path, test_size)

            qb_exec_mock.assert_called_once_with(
                'fallocate', '-l', '%sGiB' % test_size, tmp_path,
                run_as_root=self._driver._execute_as_root)

    @mock.patch.object(os, "makedirs")
    @mock.patch.object(os.path, "join", return_value="dummy_path")
    @mock.patch.object(os, "access", return_value=True)
    def test__ensure_volume_cache_ok(self, os_access_mock, os_join_mock,
                                     os_makedirs_mock):
        tmp_path = "/some/random/path"

        self._driver._ensure_volume_from_snap_cache(tmp_path)

        calls = [mock.call("dummy_path", os.F_OK),
                 mock.call("dummy_path", os.R_OK),
                 mock.call("dummy_path", os.W_OK),
                 mock.call("dummy_path", os.X_OK)]
        os_access_mock.assert_has_calls(calls)
        os_join_mock.assert_called_once_with(
            tmp_path, self._driver.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME)
        self.assertFalse(os_makedirs_mock.called)

    @mock.patch.object(fileutils, "ensure_tree")
    @mock.patch.object(os.path, "join", return_value="dummy_path")
    @mock.patch.object(os, "access", return_value=True)
    def test__ensure_volume_cache_create(self, os_access_mock, os_join_mock,
                                         os_makedirs_mock):
        tmp_path = "/some/random/path"
        os_access_mock.side_effect = [False, True, True, True]

        self._driver._ensure_volume_from_snap_cache(tmp_path)

        calls = [mock.call("dummy_path", os.F_OK),
                 mock.call("dummy_path", os.R_OK),
                 mock.call("dummy_path", os.W_OK),
                 mock.call("dummy_path", os.X_OK)]
        os_access_mock.assert_has_calls(calls)
        os_join_mock.assert_called_once_with(
            tmp_path, self._driver.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME)
        os_makedirs_mock.assert_called_once_with("dummy_path")

    @mock.patch.object(os, "makedirs")
    @mock.patch.object(os.path, "join", return_value="dummy_path")
    @mock.patch.object(os, "access", return_value=True)
    def test__ensure_volume_cache_error(self, os_access_mock, os_join_mock,
                                        os_makedirs_mock):
        tmp_path = "/some/random/path"
        os_access_mock.side_effect = [True, False, False, False]

        self.assertRaises(
            exception.VolumeDriverException,
            self._driver._ensure_volume_from_snap_cache, tmp_path)

        calls = [mock.call("dummy_path", os.F_OK),
                 mock.call("dummy_path", os.R_OK)]
        os_access_mock.assert_has_calls(calls)
        os_join_mock.assert_called_once_with(
            tmp_path, self._driver.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME)
        self.assertFalse(os_makedirs_mock.called)

    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed,
                       "_get_backing_chain_for_path")
    @ddt.data(
        [[], []],
        [[{'filename': "A"}, {'filename': CACHE_NAME}], [{'filename': "A"}]],
        [[{'filename': "A"}, {'filename': "B"}], [{'filename': "A"},
                                                  {'filename': "B"}]]
    )
    @ddt.unpack
    def test__get_backing_chain_for_path(self, test_chain,
                                         result_chain, rfs_chain_mock):
        drv = self._driver
        rfs_chain_mock.return_value = test_chain

        result = drv._get_backing_chain_for_path("foo", "bar")

        self.assertEqual(result_chain, result)

    @mock.patch.object(image_utils, 'qemu_img_info')
    @mock.patch('os.path.basename')
    def _test__qemu_img_info(self, mock_basename, mock_qemu_img_info,
                             backing_file, base_dir, valid_backing_file=True):
        drv = self._driver
        drv._execute_as_root = True
        fake_vol_name = "volume-" + self.VOLUME_UUID
        mock_info = mock_qemu_img_info.return_value
        mock_info.image = mock.sentinel.image_path
        mock_info.backing_file = backing_file

        drv._VALID_IMAGE_EXTENSIONS = ['raw', 'qcow2']

        mock_basename.side_effect = [mock.sentinel.image_basename,
                                     mock.sentinel.backing_file_basename]

        if valid_backing_file:
            img_info = drv._qemu_img_info_base(
                mock.sentinel.image_path, fake_vol_name, base_dir)
            self.assertEqual(mock_info, img_info)
            self.assertEqual(mock.sentinel.image_basename,
                             mock_info.image)
            expected_basename_calls = [mock.call(mock.sentinel.image_path)]
            if backing_file:
                self.assertEqual(mock.sentinel.backing_file_basename,
                                 mock_info.backing_file)
                expected_basename_calls.append(mock.call(backing_file))
            mock_basename.assert_has_calls(expected_basename_calls)
        else:
            self.assertRaises(exception.RemoteFSInvalidBackingFile,
                              drv._qemu_img_info_base,
                              mock.sentinel.image_path,
                              fake_vol_name, base_dir)

        mock_qemu_img_info.assert_called_with(mock.sentinel.image_path,
                                              force_share=True,
                                              run_as_root=True)

    @ddt.data(['/other_random_path', '/mnt'],
              ['/other_basedir/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID,
               '/fake_basedir'],
              ['/mnt/invalid_hash/volume-' + VOLUME_UUID, '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/invalid_vol_name', '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID + '.info',
               '/fake_basedir'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID +
               '.random-suffix', '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID +
               '.invalidext', '/mnt'])
    @ddt.unpack
    def test__qemu_img_info_invalid_backing_file(self, backing_file, basedir):
        self._test__qemu_img_info(backing_file=backing_file, base_dir=basedir,
                                  valid_backing_file=False)

    @ddt.data([None, '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID,
               '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID + '.qcow2',
               '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID +
               '.404f-404', '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/volume-' + VOLUME_UUID +
               '.tmp-snap-404f-404', '/mnt'])
    @ddt.unpack
    def test__qemu_img_info_valid_backing_file(self, backing_file, basedir):
        self._test__qemu_img_info(backing_file=backing_file, base_dir=basedir)

    @ddt.data(['/mnt/' + TEST_MNT_HASH + '/' + CACHE_NAME + '/' + VOLUME_UUID,
               '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/' + CACHE_NAME + '/' + VOLUME_UUID +
               '.child-aaaaa', '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/' + CACHE_NAME + '/' + VOLUME_UUID +
               '.parent-bbbbbb', '/mnt'],
              ['/mnt/' + TEST_MNT_HASH + '/' + CACHE_NAME + '/tmp-snap-' +
               VOLUME_UUID, '/mnt'])
    @ddt.unpack
    def test__qemu_img_info_valid_cache_backing_file(self, backing_file,
                                                     basedir):
        self._test__qemu_img_info(backing_file=backing_file, base_dir=basedir)

    @mock.patch.object(os, "listdir", return_value=["fake_vol"])
    @mock.patch.object(fileutils, "delete_if_exists")
    def test__remove_from_vol_cache_no_refs(self, fu_die_mock, os_list_mock):
        drv = self._driver
        volume = self._simple_volume()
        cache_path = drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME + "/fake_vol"
        suf = ".test_suffix"

        drv._remove_from_vol_cache(cache_path, suf, volume)

        fu_die_mock.assert_has_calls([
            mock.call(os.path.join(drv._local_volume_dir(volume),
                                   drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                                   "fake_vol.test_suffix")),
            mock.call(os.path.join(drv._local_volume_dir(volume),
                                   drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                                   "fake_vol"))])
        os_list_mock.assert_called_once_with(os.path.join(
            drv._local_volume_dir(volume),
            drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME))

    @mock.patch.object(os, "listdir", return_value=["fake_vol",
                                                    "fake_vol.more_ref"])
    @mock.patch.object(fileutils, "delete_if_exists")
    def test__remove_from_vol_cache_with_refs(self, fu_die_mock, os_list_mock):
        drv = self._driver
        volume = self._simple_volume()
        cache_path = drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME + "/fake_vol"
        suf = ".test_suffix"

        drv._remove_from_vol_cache(cache_path, suf, volume)

        fu_die_mock.assert_called_once_with(
            os.path.join(drv._local_volume_dir(volume),
                         drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                         "fake_vol.test_suffix"))
        os_list_mock.assert_called_once_with(os.path.join(
            drv._local_volume_dir(volume),
            drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME))

    def test_local_path(self):
        """local_path common use case."""
        drv = self._driver
        vol_id = self.VOLUME_UUID
        volume = self._simple_volume(_name_id=vol_id)

        self.assertEqual(
            os.path.join(self.TEST_MNT_POINT, 'volume-%s' % vol_id),
            drv.local_path(volume))

    def test_mount_quobyte_should_mount_correctly(self):
        with mock.patch.object(self._driver, '_execute') as mock_execute, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '.read_proc_mount') as mock_open, \
                mock.patch('oslo_utils.fileutils.ensure_tree') as mock_mkdir, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '._validate_volume') as mock_validate:
            # Content of /proc/mount (not mounted yet).
            mock_open.return_value = six.StringIO(
                "/dev/sda5 / ext4 rw,relatime,data=ordered 0 0")

            self._driver._mount_quobyte(self.TEST_QUOBYTE_VOLUME,
                                        self.TEST_MNT_POINT)

            mock_mkdir.assert_called_once_with(self.TEST_MNT_POINT)
            mount_call = mock.call(
                'mount.quobyte', '--disable-xattrs', self.TEST_QUOBYTE_VOLUME,
                self.TEST_MNT_POINT, run_as_root=False)

            mock_execute.assert_has_calls(
                [mount_call], any_order=False)
            mock_validate.called_once_with(self.TEST_MNT_POINT)

    def test_mount_quobyte_already_mounted_detected_seen_in_proc_mount(self):
        with mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                        '.read_proc_mount') as mock_open, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '._validate_volume') as mock_validate:
            # Content of /proc/mount (already mounted).
            mock_open.return_value = six.StringIO(
                "quobyte@%s %s fuse rw,nosuid,nodev,noatime,user_id=1000"
                ",group_id=100,default_permissions,allow_other 0 0"
                % (self.TEST_QUOBYTE_VOLUME, self.TEST_MNT_POINT))

            self._driver._mount_quobyte(self.TEST_QUOBYTE_VOLUME,
                                        self.TEST_MNT_POINT)
            mock_validate.assert_called_once_with(self.TEST_MNT_POINT)

    def test_mount_quobyte_should_suppress_already_mounted_error(self):
        """test_mount_quobyte_should_suppress_already_mounted_error

           Based on /proc/mount, the file system is not mounted yet. However,
           mount.quobyte returns with an 'already mounted' error. This is
           a last-resort safe-guard in case /proc/mount parsing was not
           successful.

           Because _mount_quobyte gets called with ensure=True, the error will
           be suppressed instead.
        """
        with mock.patch.object(self._driver, '_execute') as mock_execute, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '.read_proc_mount') as mock_open, \
                mock.patch('oslo_utils.fileutils.ensure_tree') as mock_mkdir, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '._validate_volume') as mock_validate:
            # Content of /proc/mount (empty).
            mock_open.return_value = six.StringIO()
            mock_execute.side_effect = [None, putils.ProcessExecutionError(
                stderr='is busy or already mounted')]

            self._driver._mount_quobyte(self.TEST_QUOBYTE_VOLUME,
                                        self.TEST_MNT_POINT,
                                        ensure=True)

            mock_mkdir.assert_called_once_with(self.TEST_MNT_POINT)
            mount_call = mock.call(
                'mount.quobyte', '--disable-xattrs', self.TEST_QUOBYTE_VOLUME,
                self.TEST_MNT_POINT, run_as_root=False)
            mock_execute.assert_has_calls([mount_call],
                                          any_order=False)
            mock_validate.assert_called_once_with(self.TEST_MNT_POINT)

    @mock.patch.object(psutil, "disk_partitions")
    def test_mount_quobyte_should_reraise_already_mounted_error(self,
                                                                part_mock):
        """test_mount_quobyte_should_reraise_already_mounted_error

        Like test_mount_quobyte_should_suppress_already_mounted_error
        but with ensure=False.
        """
        part_mock.return_value = []  # no quobyte@ devices
        with mock.patch.object(self._driver, '_execute') as mock_execute, \
                mock.patch('oslo_utils.fileutils.ensure_tree') as mock_mkdir, \
                mock.patch('cinder.volume.drivers.quobyte.QuobyteDriver'
                           '.read_proc_mount') as mock_open:
            mock_open.return_value = six.StringIO()
            mock_execute.side_effect = [
                None,  # mkdir
                putils.ProcessExecutionError(  # mount
                    stderr='is busy or already mounted')]

            self.assertRaises(exception.VolumeDriverException,
                              self._driver._mount_quobyte,
                              self.TEST_QUOBYTE_VOLUME,
                              self.TEST_MNT_POINT,
                              ensure=False)

            mock_mkdir.assert_called_once_with(self.TEST_MNT_POINT)
            mount_call = mock.call(
                'mount.quobyte', '--disable-xattrs', self.TEST_QUOBYTE_VOLUME,
                self.TEST_MNT_POINT, run_as_root=False)
            mock_execute.assert_has_calls([mount_call],
                                          any_order=False)

    def test_get_hash_str(self):
        """_get_hash_str should calculation correct value."""
        drv = self._driver

        self.assertEqual(self.TEST_MNT_HASH,
                         drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))

    def test_get_available_capacity_with_df(self):
        """_get_available_capacity should calculate correct value."""
        drv = self._driver

        df_total_size = 2620544
        df_avail = 1490560
        df_head = 'Filesystem 1K-blocks Used Available Use% Mounted on\n'
        df_data = 'quobyte@%s %d 996864 %d 41%% %s' % \
                  (self.TEST_QUOBYTE_VOLUME, df_total_size, df_avail,
                   self.TEST_MNT_POINT)
        df_output = df_head + df_data

        drv._get_mount_point_for_share = mock.Mock(return_value=self.
                                                   TEST_MNT_POINT)

        drv._execute = mock.Mock(return_value=(df_output, None))

        self.assertEqual((df_avail, df_total_size),
                         drv._get_available_capacity(self.TEST_QUOBYTE_VOLUME))
        (drv._get_mount_point_for_share.
            assert_called_once_with(self.TEST_QUOBYTE_VOLUME))
        (drv._execute.
         assert_called_once_with('df',
                                 '--portability',
                                 '--block-size',
                                 '1',
                                 self.TEST_MNT_POINT,
                                 run_as_root=self._driver._execute_as_root))

    def test_get_capacity_info(self):
        with mock.patch.object(self._driver, '_get_available_capacity') \
                as mock_get_available_capacity:
            drv = self._driver

            df_size = 2620544
            df_avail = 1490560

            mock_get_available_capacity.return_value = (df_avail, df_size)

            size, available, used = drv._get_capacity_info(mock.ANY)

            mock_get_available_capacity.assert_called_once_with(mock.ANY)

            self.assertEqual(df_size, size)
            self.assertEqual(df_avail, available)
            self.assertEqual(size - available, used)

    def test_load_shares_config(self):
        """_load_shares_config takes the Volume URL and strips quobyte://."""
        drv = self._driver

        drv._load_shares_config()

        self.assertIn(self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL, drv.shares)

    def test_load_shares_config_without_protocol(self):
        """Same as test_load_shares_config, but URL is without quobyte://."""
        drv = self._driver

        drv.configuration.quobyte_volume_url = \
            self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL

        drv._load_shares_config()

        self.assertIn(self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL, drv.shares)

    def test_ensure_share_mounted(self):
        """_ensure_share_mounted simple use case."""
        with mock.patch.object(self._driver, '_get_mount_point_for_share') as \
                mock_get_mount_point, \
                mock.patch.object(self._driver, '_mount_quobyte') as \
                mock_mount:
            drv = self._driver
            drv._ensure_share_mounted(self.TEST_QUOBYTE_VOLUME)

            mock_get_mount_point.assert_called_once_with(
                self.TEST_QUOBYTE_VOLUME)
            mock_mount.assert_called_once_with(
                self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL,
                mock_get_mount_point.return_value,
                ensure=True)

    def test_ensure_shares_mounted_should_save_mounting_successfully(self):
        """_ensure_shares_mounted should save share if mounted with success."""
        with mock.patch.object(self._driver, '_ensure_share_mounted') \
                as mock_ensure_share_mounted:
            drv = self._driver

            drv._ensure_shares_mounted()

            mock_ensure_share_mounted.assert_called_once_with(
                self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL)
            self.assertIn(self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL,
                          drv._mounted_shares)

    def test_ensure_shares_mounted_should_not_save_mounting_with_error(self):
        """_ensure_shares_mounted should not save if mount raised an error."""
        with mock.patch.object(self._driver, '_ensure_share_mounted') \
                as mock_ensure_share_mounted:
            drv = self._driver

            mock_ensure_share_mounted.side_effect = Exception()

            drv._ensure_shares_mounted()

            mock_ensure_share_mounted.assert_called_once_with(
                self.TEST_QUOBYTE_VOLUME_WITHOUT_PROTOCOL)
            self.assertEqual(1, len(drv.shares))
            self.assertEqual(0, len(drv._mounted_shares))

    @mock.patch.object(quobyte.QuobyteDriver, "set_nas_security_options")
    def test_do_setup(self, qb_snso_mock):
        """do_setup runs successfully."""
        drv = self._driver

        drv.do_setup(mock.create_autospec(context.RequestContext))

        qb_snso_mock.assert_called_once_with(is_new_cinder_install=mock.ANY)

    @mock.patch.object(quobyte.QuobyteDriver, "set_nas_security_options")
    def test_do_setup_overlay(self, qb_snso_mock):
        """do_setup runs successfully."""
        drv = self._driver
        drv.configuration.quobyte_qcow2_volumes = True
        drv.configuration.quobyte_overlay_volumes = True
        drv.configuration.quobyte_volume_from_snapshot_cache = True

        drv.do_setup(mock.create_autospec(context.RequestContext))

        qb_snso_mock.assert_called_once_with(is_new_cinder_install=mock.ANY)
        self.assertTrue(drv.configuration.quobyte_overlay_volumes)

    @mock.patch.object(quobyte.QuobyteDriver, "set_nas_security_options")
    def test_do_setup_no_overlay(self, qb_snso_mock):
        """do_setup runs successfully."""
        drv = self._driver
        drv.configuration.quobyte_overlay_volumes = True
        drv.configuration.quobyte_volume_from_snapshot_cache = True
        drv.configuration.quobyte_qcow2_volumes = False

        drv.do_setup(mock.create_autospec(context.RequestContext))

        qb_snso_mock.assert_called_once_with(is_new_cinder_install=mock.ANY)
        self.assertFalse(drv.configuration.quobyte_overlay_volumes)

    def test_check_for_setup_error_throws_quobyte_volume_url_not_set(self):
        """check_for_setup_error throws if 'quobyte_volume_url' is not set."""
        drv = self._driver

        drv.configuration.quobyte_volume_url = None

        self.assertRaisesAndMessageMatches(exception.VolumeDriverException,
                                           'no Quobyte volume configured',
                                           drv.check_for_setup_error)

    def test_check_for_setup_error_throws_client_not_installed(self):
        """check_for_setup_error throws if client is not installed."""
        drv = self._driver
        drv._execute = mock.Mock(side_effect=OSError
                                 (errno.ENOENT, 'No such file or directory'))

        self.assertRaisesAndMessageMatches(exception.VolumeDriverException,
                                           'mount.quobyte is not installed',
                                           drv.check_for_setup_error)
        drv._execute.assert_called_once_with('mount.quobyte',
                                             check_exit_code=False,
                                             run_as_root=False)

    def test_check_for_setup_error_throws_client_not_executable(self):
        """check_for_setup_error throws if client cannot be executed."""
        drv = self._driver

        drv._execute = mock.Mock(side_effect=OSError
                                 (errno.EPERM, 'Operation not permitted'))

        self.assertRaisesAndMessageMatches(OSError,
                                           'Operation not permitted',
                                           drv.check_for_setup_error)
        drv._execute.assert_called_once_with('mount.quobyte',
                                             check_exit_code=False,
                                             run_as_root=False)

    def test_find_share_should_throw_error_if_there_is_no_mounted_shares(self):
        """_find_share should throw error if there is no mounted share."""
        drv = self._driver

        drv._mounted_shares = []

        self.assertRaises(exception.NotFound,
                          drv._find_share,
                          self._simple_volume())

    def test_find_share(self):
        """_find_share simple use case."""
        drv = self._driver

        drv._mounted_shares = [self.TEST_QUOBYTE_VOLUME]

        self.assertEqual(self.TEST_QUOBYTE_VOLUME,
                         drv._find_share(self._simple_volume()))

    def test_find_share_does_not_throw_error_if_there_isnt_enough_space(self):
        """_find_share intentionally does not throw when no space is left."""
        with mock.patch.object(self._driver, '_get_available_capacity') \
                as mock_get_available_capacity:
            drv = self._driver

            df_size = 2620544
            df_avail = 0
            mock_get_available_capacity.return_value = (df_avail, df_size)

            drv._mounted_shares = [self.TEST_QUOBYTE_VOLUME]

            self.assertEqual(self.TEST_QUOBYTE_VOLUME,
                             drv._find_share(self._simple_volume()))

            # The current implementation does not call _get_available_capacity.
            # Future ones might do and therefore we mocked it.
            self.assertGreaterEqual(mock_get_available_capacity.call_count, 0)

    def _simple_volume(self, **kwargs):
        updates = {'id': self.VOLUME_UUID,
                   'provider_location': self.TEST_QUOBYTE_VOLUME,
                   'display_name': 'volume-%s' % self.VOLUME_UUID,
                   'name': 'volume-%s' % self.VOLUME_UUID,
                   'size': 10,
                   'status': 'available'}

        updates.update(kwargs)
        if 'display_name' not in updates:
            updates['display_name'] = 'volume-%s' % updates['id']

        return fake_volume.fake_volume_obj(self.context, **updates)

    def test_create_sparsed_volume(self):
        drv = self._driver
        volume = self._simple_volume()

        drv._create_sparsed_file = mock.Mock()
        drv._set_rw_permissions_for_all = mock.Mock()

        drv._do_create_volume(volume)
        drv._create_sparsed_file.assert_called_once_with(mock.ANY, mock.ANY)
        drv._set_rw_permissions_for_all.assert_called_once_with(mock.ANY)

    def test_create_nonsparsed_volume(self):
        drv = self._driver
        volume = self._simple_volume()

        old_value = self._configuration.quobyte_sparsed_volumes
        self._configuration.quobyte_sparsed_volumes = False

        drv._create_regular_file = mock.Mock()
        drv._set_rw_permissions_for_all = mock.Mock()

        drv._do_create_volume(volume)
        drv._create_regular_file.assert_called_once_with(mock.ANY, mock.ANY)
        drv._set_rw_permissions_for_all.assert_called_once_with(mock.ANY)

        self._configuration.quobyte_sparsed_volumes = old_value

    def test_create_qcow2_volume(self):
        drv = self._driver

        volume = self._simple_volume()
        old_value = self._configuration.quobyte_qcow2_volumes
        self._configuration.quobyte_qcow2_volumes = True

        drv._execute = mock.Mock()

        hashed = drv._get_hash_str(volume['provider_location'])
        path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                    hashed,
                                    self.VOLUME_UUID)

        drv._do_create_volume(volume)

        assert_calls = [mock.call('qemu-img', 'create', '-f', 'qcow2',
                                  '-o', 'preallocation=metadata', path,
                                  str(volume['size'] * units.Gi),
                                  run_as_root=self._driver._execute_as_root),
                        mock.call('chmod', 'ugo+rw', path,
                                  run_as_root=self._driver._execute_as_root)]
        drv._execute.assert_has_calls(assert_calls)

        self._configuration.quobyte_qcow2_volumes = old_value

    def test_create_volume_should_ensure_quobyte_mounted(self):
        """create_volume ensures shares provided in config are mounted."""
        drv = self._driver

        drv.LOG = mock.Mock()
        drv._find_share = mock.Mock()
        drv._find_share.return_value = self.TEST_QUOBYTE_VOLUME
        drv._do_create_volume = mock.Mock()
        drv._ensure_shares_mounted = mock.Mock()

        volume = self._simple_volume(size=self.TEST_SIZE_IN_GB)
        drv.create_volume(volume)

        drv._find_share.assert_called_once_with(mock.ANY)
        drv._do_create_volume.assert_called_once_with(volume)
        drv._ensure_shares_mounted.assert_called_once_with()

    def test_create_volume_should_return_provider_location(self):
        """create_volume should return provider_location with found share."""
        drv = self._driver

        drv.LOG = mock.Mock()
        drv._ensure_shares_mounted = mock.Mock()
        drv._do_create_volume = mock.Mock()
        drv._find_share = mock.Mock(return_value=self.TEST_QUOBYTE_VOLUME)

        volume = self._simple_volume(size=self.TEST_SIZE_IN_GB)
        result = drv.create_volume(volume)
        self.assertEqual(self.TEST_QUOBYTE_VOLUME, result['provider_location'])

        drv._do_create_volume.assert_called_once_with(volume)
        drv._ensure_shares_mounted.assert_called_once_with()
        drv._find_share.assert_called_once_with(volume)

    @mock.patch('oslo_utils.fileutils.delete_if_exists')
    def test_delete_volume(self, mock_delete_if_exists):
        volume = self._simple_volume()
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume_filename)
        info_file = volume_path + '.info'

        with mock.patch.object(self._driver, '_ensure_share_mounted') as \
                mock_ensure_share_mounted, \
                mock.patch.object(self._driver, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(self._driver,
                                  'get_active_image_from_info') as \
                mock_active_image_from_info, \
                mock.patch.object(self._driver, '_execute') as \
                mock_execute, \
                mock.patch.object(self._driver, '_local_path_volume') as \
                mock_local_path_volume, \
                mock.patch.object(self._driver, '_local_path_volume_info') as \
                mock_local_path_volume_info:
            self._driver._qemu_img_info = mock.Mock()
            self._driver._qemu_img_info.return_value = mock.Mock()
            self._driver._qemu_img_info.return_value.backing_file = None
            mock_local_volume_dir.return_value = self.TEST_MNT_POINT
            mock_active_image_from_info.return_value = volume_filename
            mock_local_path_volume.return_value = volume_path
            mock_local_path_volume_info.return_value = info_file

            self._driver.delete_volume(volume)

            mock_ensure_share_mounted.assert_called_once_with(
                volume['provider_location'])
            mock_local_volume_dir.assert_called_once_with(volume)
            mock_active_image_from_info.assert_called_once_with(volume)
            mock_execute.assert_called_once_with('rm', '-f', volume_path,
                                                 run_as_root=
                                                 self._driver._execute_as_root)
            mock_local_path_volume_info.assert_called_once_with(volume)
            mock_local_path_volume.assert_called_once_with(volume)
            mock_delete_if_exists.assert_any_call(volume_path)
            mock_delete_if_exists.assert_any_call(info_file)

    @mock.patch.object(os, 'access', return_value=True)
    @mock.patch('oslo_utils.fileutils.delete_if_exists')
    def test_delete_volume_backing_file(self, mock_delete_if_exists,
                                        os_acc_mock):
        drv = self._driver
        volume = self._simple_volume()
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume_filename)
        info_file = volume_path + '.info'
        drv._ensure_share_mounted = mock.Mock()
        drv._local_volume_dir = mock.Mock()
        drv._local_volume_dir.return_value = self.TEST_MNT_POINT
        drv.get_active_image_from_info = mock.Mock()
        drv.get_active_image_from_info.return_value = volume_filename
        drv._qemu_img_info = mock.Mock()
        drv._qemu_img_info.return_value = mock.Mock()
        drv._qemu_img_info.return_value.backing_file = os.path.join(
            drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME, "cached_volume_file")
        drv._remove_from_vol_cache = mock.Mock()
        drv._execute = mock.Mock()
        drv._local_path_volume = mock.Mock()
        drv._local_path_volume.return_value = volume_path
        drv._local_path_volume_info = mock.Mock()
        drv._local_path_volume_info.return_value = info_file

        drv.delete_volume(volume)

        drv._ensure_share_mounted.assert_called_once_with(
            volume['provider_location'])
        drv._local_volume_dir.assert_called_once_with(volume)
        drv.get_active_image_from_info.assert_called_once_with(volume)
        drv._qemu_img_info.assert_called_once_with(
            drv.local_path(volume), drv.get_active_image_from_info())
        drv._remove_from_vol_cache.assert_called_once_with(
            drv._qemu_img_info().backing_file, ".child-" + volume.id, volume)
        drv._execute.assert_called_once_with('rm', '-f', volume_path,
                                             run_as_root=
                                             self._driver._execute_as_root)
        drv._local_path_volume.assert_called_once_with(volume)
        drv._local_path_volume_info.assert_called_once_with(volume)
        mock_delete_if_exists.assert_any_call(volume_path)
        mock_delete_if_exists.assert_any_call(info_file)
        os_acc_mock.assert_called_once_with(drv._local_path_volume(volume),
                                            os.F_OK)

    @mock.patch.object(os, 'access', return_value=True)
    def test_delete_should_ensure_share_mounted(self, os_acc_mock):
        """delete_volume should ensure that corresponding share is mounted."""
        drv = self._driver
        drv._execute = mock.Mock()
        drv._qemu_img_info = mock.Mock()
        drv._qemu_img_info.return_value = mock.Mock()
        drv._qemu_img_info.return_value.backing_file = "/virtual/test/file"
        volume = self._simple_volume(display_name='volume-123')
        drv._ensure_share_mounted = mock.Mock()
        drv._remove_from_vol_cache = mock.Mock()

        drv.delete_volume(volume)

        (drv._ensure_share_mounted.
         assert_called_once_with(self.TEST_QUOBYTE_VOLUME))
        drv._qemu_img_info.assert_called_once_with(
            drv._local_path_volume(volume),
            drv.get_active_image_from_info(volume))
        # backing file is not in cache, no cache cleanup:
        self.assertFalse(drv._remove_from_vol_cache.called)
        drv._execute.assert_called_once_with('rm', '-f',
                                             drv.local_path(volume),
                                             run_as_root=False)
        os_acc_mock.assert_called_once_with(drv._local_path_volume(volume),
                                            os.F_OK)

    def test_delete_should_not_delete_if_provider_location_not_provided(self):
        """delete_volume shouldn't delete if provider_location missed."""
        drv = self._driver

        drv._ensure_share_mounted = mock.Mock()
        drv._execute = mock.Mock()

        volume = self._simple_volume(display_name='volume-123',
                                     provider_location=None)

        drv.delete_volume(volume)

        drv._ensure_share_mounted.assert_not_called()
        drv._execute.assert_not_called()

    @ddt.data(True, False)
    @mock.patch.object(remotefs.RemoteFSSnapDriverDistributed,
                       "_is_volume_attached")
    def test_extend_volume(self, is_attached, mock_remote_attached):
        drv = self._driver

        volume = self._simple_volume()

        volume_path = '%s/%s/volume-%s' % (self.TEST_MNT_POINT_BASE,
                                           drv._get_hash_str(
                                               self.TEST_QUOBYTE_VOLUME),
                                           self.VOLUME_UUID)

        qemu_img_info_output = """image: volume-%s
        file format: qcow2
        virtual size: 1.0G (1073741824 bytes)
        disk size: 473K
        """ % self.VOLUME_UUID

        img_info = imageutils.QemuImgInfo(qemu_img_info_output)

        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)
        self.mock_object(image_utils, 'resize_image')

        mock_remote_attached.return_value = is_attached

        if is_attached:
            self.assertRaises(exception.ExtendVolumeError, drv.extend_volume,
                              volume, 3)
        else:
            drv.extend_volume(volume, 3)

            image_utils.qemu_img_info.assert_called_once_with(volume_path,
                                                              force_share=True,
                                                              run_as_root=False
                                                              )
            image_utils.resize_image.assert_called_once_with(volume_path, 3)

    def test_copy_volume_from_snapshot(self):
        drv = self._driver

        # lots of test vars to be prepared at first
        dest_volume = self._simple_volume(
            id='c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()

        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        src_vol_path = os.path.join(vol_dir, src_volume['name'])
        dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
        info_path = os.path.join(vol_dir, src_volume['name']) + '.info'

        snapshot = self._get_fake_snapshot(src_volume)

        snap_file = dest_volume['name'] + '.' + snapshot['id']
        snap_path = os.path.join(vol_dir, snap_file)

        size = dest_volume['size']

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, src_volume['name'])
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        # mocking and testing starts here
        mock_convert = self.mock_object(image_utils, 'convert_image')
        drv._read_info_file = mock.Mock(return_value=
                                        {'active': snap_file,
                                         snapshot['id']: snap_file})
        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)
        drv._set_rw_permissions = mock.Mock()

        drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

        drv._read_info_file.assert_called_once_with(info_path)
        image_utils.qemu_img_info.assert_called_once_with(snap_path,
                                                          force_share=True,
                                                          run_as_root=False)
        (mock_convert.
         assert_called_once_with(src_vol_path,
                                 dest_vol_path,
                                 'raw',
                                 run_as_root=self._driver._execute_as_root))
        drv._set_rw_permissions.assert_called_once_with(dest_vol_path)

    @mock.patch.object(quobyte.QuobyteDriver, "_fallocate_file")
    @mock.patch.object(os, "access", return_value=True)
    def test_copy_volume_from_snapshot_cached(self, os_ac_mock,
                                              qb_falloc_mock):
        drv = self._driver
        drv.configuration.quobyte_volume_from_snapshot_cache = True

        # lots of test vars to be prepared at first
        dest_volume = self._simple_volume(
            id='c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()

        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
        info_path = os.path.join(vol_dir, src_volume['name']) + '.info'

        snapshot = self._get_fake_snapshot(src_volume)

        snap_file = dest_volume['name'] + '.' + snapshot['id']
        snap_path = os.path.join(vol_dir, snap_file)
        cache_path = os.path.join(vol_dir,
                                  drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                                  snapshot['id'])

        size = dest_volume['size']

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, src_volume['name'])
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        # mocking and testing starts here
        mock_convert = self.mock_object(image_utils, 'convert_image')
        drv._read_info_file = mock.Mock(return_value=
                                        {'active': snap_file,
                                         snapshot['id']: snap_file})
        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)
        drv._set_rw_permissions = mock.Mock()
        shutil.copyfile = mock.Mock()

        drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

        drv._read_info_file.assert_called_once_with(info_path)
        image_utils.qemu_img_info.assert_called_once_with(snap_path,
                                                          force_share=True,
                                                          run_as_root=False)
        self.assertFalse(mock_convert.called,
                         ("_convert_image was called but should not have been")
                         )
        os_ac_mock.assert_called_once_with(
            drv._local_volume_from_snap_cache_path(snapshot), os.F_OK)
        qb_falloc_mock.assert_called_once_with(dest_vol_path, size)
        shutil.copyfile.assert_called_once_with(cache_path, dest_vol_path)
        drv._set_rw_permissions.assert_called_once_with(dest_vol_path)

    @mock.patch.object(os, "symlink")
    @mock.patch.object(os, "access", return_value=False)
    def test_copy_volume_from_snapshot_not_cached_overlay(self, os_ac_mock,
                                                          os_sl_mock):
        drv = self._driver
        drv.configuration.quobyte_qcow2_volumes = True
        drv.configuration.quobyte_volume_from_snapshot_cache = True
        drv.configuration.quobyte_overlay_volumes = True

        # lots of test vars to be prepared at first
        dest_volume = self._simple_volume(
            id='c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()
        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        src_vol_path = os.path.join(vol_dir, src_volume['name'])

        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
        info_path = os.path.join(vol_dir, src_volume['name']) + '.info'

        snapshot = self._get_fake_snapshot(src_volume)

        snap_file = dest_volume['name'] + '.' + snapshot['id']
        snap_path = os.path.join(vol_dir, snap_file)

        size = dest_volume['size']

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, src_volume['name'])
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        # mocking and testing starts here
        mock_convert = self.mock_object(image_utils, 'convert_image')
        drv._read_info_file = mock.Mock(return_value=
                                        {'active': snap_file,
                                         snapshot['id']: snap_file})
        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)
        drv._set_rw_permissions = mock.Mock()
        drv._create_overlay_volume_from_snapshot = mock.Mock()

        drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

        drv._read_info_file.assert_called_once_with(info_path)
        os_ac_mock.assert_called_once_with(
            drv._local_volume_from_snap_cache_path(snapshot), os.F_OK)
        image_utils.qemu_img_info.assert_called_once_with(snap_path,
                                                          force_share=True,
                                                          run_as_root=False)
        (mock_convert.
         assert_called_once_with(
             src_vol_path,
             drv._local_volume_from_snap_cache_path(snapshot), 'qcow2',
             run_as_root=self._driver._execute_as_root))
        os_sl_mock.assert_called_once_with(
            src_vol_path,
            drv._local_volume_from_snap_cache_path(snapshot) + '.parent-'
            + snapshot.id)
        drv._create_overlay_volume_from_snapshot.assert_called_once_with(
            dest_volume, snapshot, size, 'qcow2')
        drv._set_rw_permissions.assert_called_once_with(dest_vol_path)

    @mock.patch.object(quobyte.QuobyteDriver, "_fallocate_file")
    def test_copy_volume_from_snapshot_not_cached(self, qb_falloc_mock):
        drv = self._driver
        drv.configuration.quobyte_volume_from_snapshot_cache = True

        # lots of test vars to be prepared at first
        dest_volume = self._simple_volume(
            id='c1073000-0000-0000-0000-0000000c1073')
        src_volume = self._simple_volume()

        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        src_vol_path = os.path.join(vol_dir, src_volume['name'])
        dest_vol_path = os.path.join(vol_dir, dest_volume['name'])
        info_path = os.path.join(vol_dir, src_volume['name']) + '.info'

        snapshot = self._get_fake_snapshot(src_volume)

        snap_file = dest_volume['name'] + '.' + snapshot['id']
        snap_path = os.path.join(vol_dir, snap_file)
        cache_path = os.path.join(vol_dir,
                                  drv.QUOBYTE_VOLUME_SNAP_CACHE_DIR_NAME,
                                  snapshot['id'])

        size = dest_volume['size']

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        backing file: %s
        """ % (snap_file, src_volume['name'])
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        # mocking and testing starts here
        mock_convert = self.mock_object(image_utils, 'convert_image')
        drv._read_info_file = mock.Mock(return_value=
                                        {'active': snap_file,
                                         snapshot['id']: snap_file})
        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)
        drv._set_rw_permissions = mock.Mock()
        self.mock_object(shutil, 'copyfile')

        drv._copy_volume_from_snapshot(snapshot, dest_volume, size)

        drv._read_info_file.assert_called_once_with(info_path)
        image_utils.qemu_img_info.assert_called_once_with(snap_path,
                                                          force_share=True,
                                                          run_as_root=False)
        (mock_convert.
         assert_called_once_with(
             src_vol_path,
             drv._local_volume_from_snap_cache_path(snapshot), 'raw',
             run_as_root=self._driver._execute_as_root))
        qb_falloc_mock.assert_called_once_with(dest_vol_path, size)
        shutil.copyfile.assert_called_once_with(cache_path, dest_vol_path)
        drv._set_rw_permissions.assert_called_once_with(dest_vol_path)

    @ddt.data(['available', True], ['backing-up', True],
              ['creating', False], ['deleting', False])
    @ddt.unpack
    def test_create_volume_from_snapshot(self, state, should_work):
        drv = self._driver

        src_volume = self._simple_volume()

        snap_ref = fake_snapshot.fake_snapshot_obj(
            self.context,
            volume_name=src_volume.name,
            display_name='clone-snap-%s' % src_volume.id,
            volume_size=src_volume.size,
            volume_id=src_volume.id,
            id=self.SNAP_UUID,
            status=state)
        snap_ref.volume = src_volume

        new_volume = self._simple_volume(size=snap_ref.volume_size)

        drv._ensure_shares_mounted = mock.Mock()
        drv._find_share = mock.Mock(return_value=self.TEST_QUOBYTE_VOLUME)
        drv._copy_volume_from_snapshot = mock.Mock()

        if should_work:
            drv.create_volume_from_snapshot(new_volume, snap_ref)

            drv._ensure_shares_mounted.assert_called_once_with()
            drv._find_share.assert_called_once_with(new_volume)
            (drv._copy_volume_from_snapshot.
             assert_called_once_with(snap_ref, new_volume, new_volume['size']))
        else:
            self.assertRaises(exception.InvalidSnapshot,
                              drv.create_volume_from_snapshot,
                              new_volume,
                              snap_ref)

    def test_initialize_connection(self):
        drv = self._driver

        volume = self._simple_volume()
        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(self.TEST_QUOBYTE_VOLUME))
        vol_path = os.path.join(vol_dir, volume['name'])

        qemu_img_output = """image: %s
        file format: raw
        virtual size: 1.0G (1073741824 bytes)
        disk size: 173K
        """ % volume['name']
        img_info = imageutils.QemuImgInfo(qemu_img_output)

        drv.get_active_image_from_info = mock.Mock(return_value=volume['name'])
        self.mock_object(image_utils, 'qemu_img_info', return_value=img_info)

        conn_info = drv.initialize_connection(volume, None)

        drv.get_active_image_from_info.assert_called_once_with(volume)
        image_utils.qemu_img_info.assert_called_once_with(vol_path,
                                                          force_share=True,
                                                          run_as_root=False)

        self.assertEqual('raw', conn_info['data']['format'])
        self.assertEqual('quobyte', conn_info['driver_volume_type'])
        self.assertEqual(volume['name'], conn_info['data']['name'])
        self.assertEqual(self.TEST_MNT_POINT_BASE,
                         conn_info['mount_point_base'])

    @mock.patch('cinder.db.volume_glance_metadata_get', return_value={})
    def test_copy_volume_to_image_raw_image(self, vol_glance_metadata):
        drv = self._driver

        volume_type_id = db.volume_type_create(
            self.context, {'name': 'quo_type', 'extra_specs': {}}).get('id')
        volume = self._simple_volume(volume_type_id=volume_type_id)
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume['name'])
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: %s
            file format: raw
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume['name']
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = volume_path

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_once_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path,
                                                       force_share=True,
                                                       run_as_root=False)
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path, run_as_root=False,
                store_id=None, base_image_ref=None, compress=True,
                volume_format='raw')
            self.assertTrue(mock_create_temporary_file.called)

    @mock.patch('cinder.db.volume_glance_metadata_get', return_value={})
    def test_copy_volume_to_image_qcow2_image(self, vol_glance_metadata):
        """Upload a qcow2 image file which has to be converted to raw first."""
        drv = self._driver

        volume_type_id = db.volume_type_create(
            self.context, {'name': 'quo_type', 'extra_specs': {}}).get('id')
        volume = self._simple_volume(volume_type_id=volume_type_id)
        volume_path = '%s/%s' % (self.TEST_MNT_POINT, volume['name'])
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'convert_image') as \
                mock_convert_image, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: %s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            """ % volume['name']
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = self.TEST_TMP_FILE

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path,
                                                       force_share=True,
                                                       run_as_root=False)
            mock_convert_image.assert_called_once_with(
                volume_path, upload_path, 'raw', run_as_root=False)
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path, run_as_root=False,
                store_id=None, base_image_ref=None, compress=True,
                volume_format='raw')
            self.assertTrue(mock_create_temporary_file.called)

    @mock.patch('cinder.db.volume_glance_metadata_get', return_value={})
    def test_copy_volume_to_image_snapshot_exists(self, vol_glance_metadata):
        """Upload an active snapshot which has to be converted to raw first."""
        drv = self._driver

        volume_type_id = db.volume_type_create(
            self.context, {'name': 'quo_type', 'extra_specs': {}}).get('id')
        volume = self._simple_volume(volume_type_id=volume_type_id)
        volume_path = '%s/volume-%s' % (self.TEST_MNT_POINT, self.VOLUME_UUID)
        volume_filename = 'volume-%s' % self.VOLUME_UUID
        image_meta = {'id': '10958016-e196-42e3-9e7f-5d8927ae3099'}

        with mock.patch.object(drv, 'get_active_image_from_info') as \
                mock_get_active_image_from_info, \
                mock.patch.object(drv, '_local_volume_dir') as \
                mock_local_volume_dir, \
                mock.patch.object(image_utils, 'qemu_img_info') as \
                mock_qemu_img_info, \
                mock.patch.object(image_utils, 'convert_image') as \
                mock_convert_image, \
                mock.patch.object(image_utils, 'upload_volume') as \
                mock_upload_volume, \
                mock.patch.object(image_utils, 'create_temporary_file') as \
                mock_create_temporary_file:
            mock_get_active_image_from_info.return_value = volume['name']

            mock_local_volume_dir.return_value = self.TEST_MNT_POINT

            mock_create_temporary_file.return_value = self.TEST_TMP_FILE

            qemu_img_output = """image: volume-%s.%s
            file format: qcow2
            virtual size: 1.0G (1073741824 bytes)
            disk size: 173K
            backing file: %s
            """ % (self.VOLUME_UUID, self.SNAP_UUID, volume_filename)
            img_info = imageutils.QemuImgInfo(qemu_img_output)
            mock_qemu_img_info.return_value = img_info

            upload_path = self.TEST_TMP_FILE

            drv.copy_volume_to_image(mock.ANY, volume, mock.ANY, image_meta)

            mock_get_active_image_from_info.assert_called_once_with(volume)
            mock_local_volume_dir.assert_called_with(volume)
            mock_qemu_img_info.assert_called_once_with(volume_path,
                                                       force_share=True,
                                                       run_as_root=False)
            mock_convert_image.assert_called_once_with(
                volume_path, upload_path, 'raw', run_as_root=False)
            mock_upload_volume.assert_called_once_with(
                mock.ANY, mock.ANY, mock.ANY, upload_path, run_as_root=False,
                store_id=None, base_image_ref=None, compress=True,
                volume_format='raw')
            self.assertTrue(mock_create_temporary_file.called)

    def test_set_nas_security_options_default(self):
        drv = self._driver
        self.assertEqual("true", drv.configuration.nas_secure_file_operations)
        self.assertEqual("true",
                         drv.configuration.nas_secure_file_permissions)
        self.assertFalse(drv._execute_as_root)

    def test_set_nas_security_options_insecure(self):
        drv = self._driver
        drv.configuration.nas_secure_file_operations = "false"
        drv.configuration.nas_secure_file_permissions = "false"

        drv.set_nas_security_options(is_new_cinder_install=True)

        self.assertEqual("false",
                         drv.configuration.nas_secure_file_operations)
        self.assertEqual("false",
                         drv.configuration.nas_secure_file_permissions)
        self.assertTrue(drv._execute_as_root)

    def test_set_nas_security_options_explicitly_secure(self):
        drv = self._driver
        drv.configuration.nas_secure_file_operations = "true"
        drv.configuration.nas_secure_file_permissions = "true"

        drv.set_nas_security_options(is_new_cinder_install=True)

        self.assertEqual("true",
                         drv.configuration.nas_secure_file_operations)
        self.assertEqual("true",
                         drv.configuration.nas_secure_file_permissions)
        self.assertFalse(drv._execute_as_root)

    @mock.patch.object(psutil, "disk_partitions")
    @mock.patch.object(os, "stat")
    def test_validate_volume_all_good_prefix_val(self, stat_mock, part_mock):
        part_mock.return_value = self.get_mock_partitions()
        drv = self._driver

        def statMockCall(*args):
            if args[0] == self.TEST_MNT_POINT:
                stat_result = mock.Mock()
                stat_result.st_size = 0
                return stat_result
            return os.stat(args)
        stat_mock.side_effect = statMockCall

        drv._validate_volume(self.TEST_MNT_POINT)

        stat_mock.assert_called_once_with(self.TEST_MNT_POINT)
        part_mock.assert_called_once_with(all=True)

    @mock.patch.object(psutil, "disk_partitions")
    @mock.patch.object(os, "stat")
    def test_validate_volume_all_good_subtype_val(self, stat_mock, part_mock):
        part_mock.return_value = self.get_mock_partitions()
        part_mock.return_value[0].device = "not_quobyte"
        part_mock.return_value[0].fstype = "fuse.quobyte"
        drv = self._driver

        def statMockCall(*args):
            if args[0] == self.TEST_MNT_POINT:
                stat_result = mock.Mock()
                stat_result.st_size = 0
                return stat_result
            return os.stat(args)
        stat_mock.side_effect = statMockCall

        drv._validate_volume(self.TEST_MNT_POINT)

        stat_mock.assert_called_once_with(self.TEST_MNT_POINT)
        part_mock.assert_called_once_with(all=True)

    @mock.patch.object(psutil, "disk_partitions")
    @mock.patch.object(os, "stat")
    def test_validate_volume_mount_not_working(self, stat_mock, part_mock):
        part_mock.return_value = self.get_mock_partitions()
        drv = self._driver

        def statMockCall(*args):
            if args[0] == self.TEST_MNT_POINT:
                raise exception.VolumeDriverException()
        stat_mock.side_effect = [statMockCall, os.stat]

        self.assertRaises(
            exception.VolumeDriverException,
            drv._validate_volume,
            self.TEST_MNT_POINT)
        stat_mock.assert_called_once_with(self.TEST_MNT_POINT)
        part_mock.assert_called_once_with(all=True)

    @mock.patch.object(psutil, "disk_partitions")
    def test_validate_volume_no_mtab_entry(self, part_mock):
        part_mock.return_value = []  # no quobyte@ devices
        msg = ("Volume driver reported an error: "
               "No matching Quobyte mount entry for %(mpt)s"
               " could be found for validation in partition list."
               % {'mpt': self.TEST_MNT_POINT})

        self.assertRaisesAndMessageMatches(
            exception.VolumeDriverException,
            msg,
            self._driver._validate_volume,
            self.TEST_MNT_POINT)

    @mock.patch.object(psutil, "disk_partitions")
    def test_validate_volume_wrong_mount_type(self, part_mock):
        mypart = mock.Mock()
        mypart.device = "not-quobyte"
        mypart.mountpoint = self.TEST_MNT_POINT
        part_mock.return_value = [mypart]
        msg = ("Volume driver reported an error: "
               "The mount %(mpt)s is not a valid"
               " Quobyte volume according to partition list."
               % {'mpt': self.TEST_MNT_POINT})
        drv = self._driver

        self.assertRaisesAndMessageMatches(
            exception.VolumeDriverException,
            msg,
            drv._validate_volume,
            self.TEST_MNT_POINT)
        part_mock.assert_called_once_with(all=True)

    @mock.patch.object(psutil, "disk_partitions")
    def test_validate_volume_stale_mount(self, part_mock):
        part_mock.return_value = self.get_mock_partitions()
        drv = self._driver

        # As this uses a local fs the dir size is >0, raising an exception
        self.assertRaises(
            exception.VolumeDriverException,
            drv._validate_volume,
            self.TEST_MNT_POINT)



# Stubs for oslo_i18n._factory (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

class TranslatorFactory:
    domain: Any = ...
    localedir: Any = ...
    def __init__(self, domain: Any, localedir: Optional[Any] = ...) -> None: ...
    @property
    def primary(self) -> ty.Callable[str]: ...
    @property
    def contextual_form(self): ...
    @property
    def plural_form(self): ...
    @property
    def log_info(self): ...
    @property
    def log_warning(self): ...
    @property
    def log_error(self): ...
    @property
    def log_critical(self): ...



# Copyright (C) 2021 NEC corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Unit tests for NEC Driver."""

from unittest import mock

from oslo_config import cfg
import requests

from cinder import context as cinder_context
from cinder import db
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_rest_api
from cinder.volume.drivers.nec.v import nec_v_fc
from cinder.volume.drivers.nec.v import nec_v_rest
from cinder.volume import volume_utils

# Configuration parameter values
CONFIG_MAP = {
    'serial': '886000123456',
    'my_ip': '127.0.0.1',
    'rest_server_ip_addr': '172.16.18.108',
    'rest_server_ip_port': '23451',
    'port_id': 'CL1-A',
    'host_grp_name': 'NEC-0123456789abcdef',
    'host_mode': 'LINUX/IRIX',
    'host_wwn': '0123456789abcdef',
    'target_wwn': '1111111123456789',
    'user_id': 'user',
    'user_pass': 'password',
    'pool_name': 'test_pool',
    'auth_user': 'auth_user',
    'auth_password': 'auth_password',
}

# Dummy response
DEFAULT_CONNECTOR = {
    'host': 'host',
    'ip': CONFIG_MAP['my_ip'],
    'wwpns': [CONFIG_MAP['host_wwn']],
    'multipath': False,
}

# Dummy response for REST API
POST_SESSIONS_RESULT = {
    "token": "b74777a3-f9f0-4ea8-bd8f-09847fac48d3",
    "sessionId": 0,
}

GET_PORTS_RESULT = {
    "data": [
        {
            "portId": CONFIG_MAP['port_id'],
            "portType": "FIBRE",
            "portAttributes": [
                "TAR",
                "MCU",
                "RCU",
                "ELUN"
            ],
            "fabricMode": True,
            "portConnection": "PtoP",
            "lunSecuritySetting": True,
            "wwn": CONFIG_MAP['target_wwn'],
        },
    ],
}

GET_HOST_WWNS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "hostWwn": CONFIG_MAP['host_wwn'],
        },
    ],
}

COMPLETED_SUCCEEDED_RESULT = {
    "status": "Completed",
    "state": "Succeeded",
    "affectedResources": ('a/b/c/1',),
}


def _brick_get_connector_properties(multipath=False, enforce_multipath=False):
    """Return a predefined connector object."""
    return DEFAULT_CONNECTOR


class FakeResponse():

    def __init__(self, status_code, data=None, headers=None):
        self.status_code = status_code
        self.data = data
        self.text = data
        self.content = data
        self.headers = {'Content-Type': 'json'} if headers is None else headers

    def json(self):
        return self.data


class VStorageRESTFCDriverTest(test.TestCase):
    """Unit test class for NEC REST interface fibre channel module."""

    test_existing_ref = {'source-id': '1'}
    test_existing_ref_name = {
        'source-name': '15960cc7-38c9-4c5b-b4f1-365be5eeed45'}

    def setUp(self):
        """Set up the test environment."""
        def _set_required(opts, required):
            for opt in opts:
                opt.required = required

        # Initialize Cinder and avoid checking driver options.
        rest_required_opts = [
            opt for opt in nec_v_rest.REST_VOLUME_OPTS if opt.required]
        common_required_opts = [
            opt for opt in nec_v_rest.COMMON_VOLUME_OPTS if opt.required]
        _set_required(rest_required_opts, False)
        _set_required(common_required_opts, False)
        super(VStorageRESTFCDriverTest, self).setUp()
        _set_required(rest_required_opts, True)
        _set_required(common_required_opts, True)

        self.configuration = mock.Mock(conf.Configuration)
        self.ctxt = cinder_context.get_admin_context()
        self._setup_config()
        self._setup_driver()

    def _setup_config(self):
        """Set configuration parameter values."""
        self.configuration.config_group = "REST"

        self.configuration.volume_backend_name = "RESTFC"
        self.configuration.volume_driver = (
            "cinder.volume.drivers.nec.v.nec_v_fc.VStorageFCDriver")
        self.configuration.reserved_percentage = "0"
        self.configuration.use_multipath_for_image_xfer = False
        self.configuration.enforce_multipath_for_image_xfer = False
        self.configuration.max_over_subscription_ratio = 500.0
        self.configuration.driver_ssl_cert_verify = False

        self.configuration.nec_v_storage_id = CONFIG_MAP['serial']
        self.configuration.nec_v_pool = "30"
        self.configuration.nec_v_snap_pool = None
        self.configuration.nec_v_ldev_range = "0-1"
        self.configuration.nec_v_target_ports = [CONFIG_MAP['port_id']]
        self.configuration.nec_v_compute_target_ports = [
            CONFIG_MAP['port_id']]
        self.configuration.nec_v_group_create = True
        self.configuration.nec_v_group_delete = True
        self.configuration.nec_v_copy_speed = 3
        self.configuration.nec_v_copy_check_interval = 3
        self.configuration.nec_v_async_copy_check_interval = 10

        self.configuration.san_login = CONFIG_MAP['user_id']
        self.configuration.san_password = CONFIG_MAP['user_pass']
        self.configuration.san_ip = CONFIG_MAP[
            'rest_server_ip_addr']
        self.configuration.san_api_port = CONFIG_MAP[
            'rest_server_ip_port']
        self.configuration.nec_v_rest_disable_io_wait = True
        self.configuration.nec_v_rest_tcp_keepalive = True
        self.configuration.nec_v_discard_zero_page = True
        self.configuration.nec_v_rest_number = "0"
        self.configuration.nec_v_lun_timeout = hbsd_rest._LUN_TIMEOUT
        self.configuration.nec_v_lun_retry_interval = (
            hbsd_rest._LUN_RETRY_INTERVAL)
        self.configuration.nec_v_restore_timeout = hbsd_rest._RESTORE_TIMEOUT
        self.configuration.nec_v_state_transition_timeout = (
            hbsd_rest._STATE_TRANSITION_TIMEOUT)
        self.configuration.nec_v_lock_timeout = hbsd_rest_api._LOCK_TIMEOUT
        self.configuration.nec_v_rest_timeout = hbsd_rest_api._REST_TIMEOUT
        self.configuration.nec_v_extend_timeout = (
            hbsd_rest_api._EXTEND_TIMEOUT)
        self.configuration.nec_v_exec_retry_interval = (
            hbsd_rest_api._EXEC_RETRY_INTERVAL)
        self.configuration.nec_v_rest_connect_timeout = (
            hbsd_rest_api._DEFAULT_CONNECT_TIMEOUT)
        self.configuration.nec_v_rest_job_api_response_timeout = (
            hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT)
        self.configuration.nec_v_rest_get_api_response_timeout = (
            hbsd_rest_api._GET_API_RESPONSE_TIMEOUT)
        self.configuration.nec_v_rest_server_busy_timeout = (
            hbsd_rest_api._REST_SERVER_BUSY_TIMEOUT)
        self.configuration.nec_v_rest_keep_session_loop_interval = (
            hbsd_rest_api._KEEP_SESSION_LOOP_INTERVAL)
        self.configuration.nec_v_rest_another_ldev_mapped_retry_timeout = (
            hbsd_rest_api._ANOTHER_LDEV_MAPPED_RETRY_TIMEOUT)
        self.configuration.nec_v_rest_tcp_keepidle = (
            hbsd_rest_api._TCP_KEEPIDLE)
        self.configuration.nec_v_rest_tcp_keepintvl = (
            hbsd_rest_api._TCP_KEEPINTVL)
        self.configuration.nec_v_rest_tcp_keepcnt = (
            hbsd_rest_api._TCP_KEEPCNT)
        self.configuration.nec_v_host_mode_options = []

        self.configuration.nec_v_zoning_request = False

        self.configuration.san_thin_provision = True
        self.configuration.san_private_key = ''
        self.configuration.san_clustername = ''
        self.configuration.san_ssh_port = '22'
        self.configuration.san_is_local = False
        self.configuration.ssh_conn_timeout = '30'
        self.configuration.ssh_min_pool_conn = '1'
        self.configuration.ssh_max_pool_conn = '5'

        self.configuration.use_chap_auth = True
        self.configuration.chap_username = CONFIG_MAP['auth_user']
        self.configuration.chap_password = CONFIG_MAP['auth_password']

        self.configuration.safe_get = self._fake_safe_get

        CONF = cfg.CONF
        CONF.my_ip = CONFIG_MAP['my_ip']

    def _fake_safe_get(self, value):
        """Retrieve a configuration value avoiding throwing an exception."""
        try:
            val = getattr(self.configuration, value)
        except AttributeError:
            val = None
        return val

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def _setup_driver(
            self, brick_get_connector_properties=None, request=None):
        """Set up the driver environment."""
        self.driver = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                               FakeResponse(200, GET_PORTS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
        self.driver.do_setup(None)
        self.driver.check_for_setup_error()
        self.driver.local_path(None)
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.client.keep_session_loop.stop()

    def tearDown(self):
        self.client = None
        super(VStorageRESTFCDriverTest, self).tearDown()

    # API test cases
    def test_configuration(self):
        drv = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        self.assertEqual(drv.configuration.hitachi_storage_id,
                         drv.configuration.nec_v_storage_id)
        self.assertEqual(drv.configuration.hitachi_pool,
                         drv.configuration.nec_v_pool)
        self.assertEqual(drv.configuration.hitachi_snap_pool,
                         drv.configuration.nec_v_snap_pool)
        self.assertEqual(drv.configuration.hitachi_ldev_range,
                         drv.configuration.nec_v_ldev_range)
        self.assertEqual(drv.configuration.hitachi_target_ports,
                         drv.configuration.nec_v_target_ports)
        self.assertEqual(drv.configuration.hitachi_compute_target_ports,
                         drv.configuration.nec_v_compute_target_ports)
        self.assertEqual(drv.configuration.hitachi_group_create,
                         drv.configuration.nec_v_group_create)
        self.assertEqual(drv.configuration.hitachi_group_delete,
                         drv.configuration.nec_v_group_delete)
        self.assertEqual(drv.configuration.hitachi_copy_speed,
                         drv.configuration.nec_v_copy_speed)
        self.assertEqual(drv.configuration.hitachi_copy_check_interval,
                         drv.configuration.nec_v_copy_check_interval)
        self.assertEqual(drv.configuration.hitachi_async_copy_check_interval,
                         drv.configuration.nec_v_async_copy_check_interval)
        self.assertEqual(drv.configuration.hitachi_rest_disable_io_wait,
                         drv.configuration.nec_v_rest_disable_io_wait)
        self.assertEqual(drv.configuration.hitachi_rest_tcp_keepalive,
                         drv.configuration.nec_v_rest_tcp_keepalive)
        self.assertEqual(drv.configuration.hitachi_discard_zero_page,
                         drv.configuration.nec_v_discard_zero_page)
        self.assertEqual(drv.configuration.hitachi_lun_timeout,
                         drv.configuration.nec_v_lun_timeout)
        self.assertEqual(drv.configuration.hitachi_lun_retry_interval,
                         drv.configuration.nec_v_lun_retry_interval)
        self.assertEqual(drv.configuration.hitachi_restore_timeout,
                         drv.configuration.nec_v_restore_timeout)
        self.assertEqual(drv.configuration.hitachi_state_transition_timeout,
                         drv.configuration.nec_v_state_transition_timeout)
        self.assertEqual(drv.configuration.hitachi_lock_timeout,
                         drv.configuration.nec_v_lock_timeout)
        self.assertEqual(drv.configuration.hitachi_rest_timeout,
                         drv.configuration.nec_v_rest_timeout)
        self.assertEqual(drv.configuration.hitachi_extend_timeout,
                         drv.configuration.nec_v_extend_timeout)
        self.assertEqual(drv.configuration.hitachi_exec_retry_interval,
                         drv.configuration.nec_v_exec_retry_interval)
        self.assertEqual(drv.configuration.hitachi_rest_connect_timeout,
                         drv.configuration.nec_v_rest_connect_timeout)
        self.assertEqual(
            drv.configuration.hitachi_rest_job_api_response_timeout,
            drv.configuration.nec_v_rest_job_api_response_timeout)
        self.assertEqual(
            drv.configuration.hitachi_rest_get_api_response_timeout,
            drv.configuration.nec_v_rest_get_api_response_timeout)
        self.assertEqual(drv.configuration.hitachi_rest_server_busy_timeout,
                         drv.configuration.nec_v_rest_server_busy_timeout)
        self.assertEqual(
            drv.configuration.hitachi_rest_keep_session_loop_interval,
            drv.configuration.nec_v_rest_keep_session_loop_interval)
        self.assertEqual(
            drv.configuration.hitachi_rest_another_ldev_mapped_retry_timeout,
            drv.configuration.nec_v_rest_another_ldev_mapped_retry_timeout)
        self.assertEqual(drv.configuration.hitachi_rest_tcp_keepidle,
                         drv.configuration.nec_v_rest_tcp_keepidle)
        self.assertEqual(drv.configuration.hitachi_rest_tcp_keepintvl,
                         drv.configuration.nec_v_rest_tcp_keepintvl)
        self.assertEqual(drv.configuration.hitachi_rest_tcp_keepcnt,
                         drv.configuration.nec_v_rest_tcp_keepcnt)
        self.assertEqual(drv.configuration.hitachi_host_mode_options,
                         drv.configuration.nec_v_host_mode_options)
        self.assertEqual(drv.configuration.hitachi_zoning_request,
                         drv.configuration.nec_v_zoning_request)

    def test_driverinfo(self):
        drv = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        self.assertEqual(drv.common.driver_info['version'],
                         "1.0.0")
        self.assertEqual(drv.common.driver_info['proto'],
                         "FC")
        self.assertEqual(drv.common.driver_info['hba_id'],
                         "wwpns")
        self.assertEqual(drv.common.driver_info['hba_id_type'],
                         "World Wide Name")
        self.assertEqual(drv.common.driver_info['msg_id']['target'].msg_id,
                         308)
        self.assertEqual(drv.common.driver_info['volume_backend_name'],
                         "NECFC")
        self.assertEqual(drv.common.driver_info['volume_type'],
                         "fibre_channel")
        self.assertEqual(drv.common.driver_info['param_prefix'],
                         "nec_v")
        self.assertEqual(drv.common.driver_info['vendor_name'],
                         "NEC")
        self.assertEqual(drv.common.driver_info['driver_prefix'],
                         "NEC")
        self.assertEqual(drv.common.driver_info['driver_file_prefix'],
                         "nec")
        self.assertEqual(drv.common.driver_info['target_prefix'],
                         "NEC-")
        self.assertEqual(drv.common.driver_info['hdp_vol_attr'],
                         "DP")
        self.assertEqual(drv.common.driver_info['hdt_vol_attr'],
                         "DT")
        self.assertEqual(drv.common.driver_info['nvol_ldev_type'],
                         "DP-VOL")
        self.assertEqual(drv.common.driver_info['target_iqn_suffix'],
                         ".nec-target")
        self.assertEqual(drv.common.driver_info['pair_attr'],
                         "SS")

    @mock.patch.object(requests.Session, "request")
    def test_create_volume(self, request):
        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
        ret = self.driver.create_volume(fake_volume.fake_volume_obj(self.ctxt))
        self.assertEqual('1', ret['provider_location'])
        self.assertEqual(2, request.call_count)



# Copyright (C) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for transfers table."""

from cinder import context
from cinder import db
from cinder.db.sqlalchemy import api as db_api
from cinder.db.sqlalchemy import models
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils


class TransfersTableTestCase(test.TestCase):
    """Test case for transfers model."""

    def setUp(self):
        super(TransfersTableTestCase, self).setUp()
        self.ctxt = context.RequestContext(user_id=fake.USER_ID,
                                           project_id=fake.PROJECT_ID)

    def _create_transfer(self, volume_id=None, source_project_id=None):
        """Create a transfer object."""
        transfer = {'display_name': 'display_name',
                    'salt': 'salt',
                    'crypt_hash': 'crypt_hash'}
        if volume_id is not None:
            transfer['volume_id'] = volume_id
        if source_project_id is not None:
            transfer['source_project_id'] = source_project_id
        return db.transfer_create(self.ctxt, transfer)['id']

    def test_transfer_create(self):
        # If the volume_id is Null a KeyError exception will be raised.
        self.assertRaises(KeyError,
                          self._create_transfer)

        volume_id = utils.create_volume(self.ctxt)['id']
        self._create_transfer(volume_id)

    def test_transfer_create_not_available(self):
        volume_id = utils.create_volume(self.ctxt, size=1,
                                        status='notavailable')['id']
        self.assertRaises(exception.InvalidVolume,
                          self._create_transfer,
                          volume_id)

    def test_transfer_get(self):
        volume_id1 = utils.create_volume(self.ctxt)['id']
        xfer_id1 = self._create_transfer(volume_id1)

        xfer = db.transfer_get(self.ctxt, xfer_id1)
        self.assertEqual(volume_id1, xfer.volume_id, "Unexpected volume_id")

        nctxt = context.RequestContext(user_id='new_user_id',
                                       project_id='new_project_id')
        self.assertRaises(exception.TransferNotFound,
                          db.transfer_get, nctxt, xfer_id1)

        xfer = db.transfer_get(nctxt.elevated(), xfer_id1)
        self.assertEqual(volume_id1, xfer.volume_id, "Unexpected volume_id")

    def test_transfer_get_all(self):
        volume_id1 = utils.create_volume(self.ctxt)['id']
        volume_id2 = utils.create_volume(self.ctxt)['id']
        self._create_transfer(volume_id1)
        self._create_transfer(volume_id2)

        self.assertRaises(exception.NotAuthorized,
                          db.transfer_get_all,
                          self.ctxt)
        xfer = db.transfer_get_all(context.get_admin_context())
        self.assertEqual(2, len(xfer), "Unexpected number of transfer records")

        xfer = db.transfer_get_all_by_project(self.ctxt, self.ctxt.project_id)
        self.assertEqual(2, len(xfer), "Unexpected number of transfer records")

        nctxt = context.RequestContext(user_id=fake.USER2_ID,
                                       project_id=fake.PROJECT2_ID)
        self.assertRaises(exception.NotAuthorized,
                          db.transfer_get_all_by_project,
                          nctxt, self.ctxt.project_id)
        xfer = db.transfer_get_all_by_project(nctxt.elevated(),
                                              self.ctxt.project_id)
        self.assertEqual(2, len(xfer), "Unexpected number of transfer records")

    def test_transfer_destroy(self):
        volume_id = utils.create_volume(self.ctxt)['id']
        volume_id2 = utils.create_volume(self.ctxt)['id']
        xfer_id1 = self._create_transfer(volume_id)
        xfer_id2 = self._create_transfer(volume_id2)

        xfer = db.transfer_get_all(context.get_admin_context())
        self.assertEqual(2, len(xfer), "Unexpected number of transfer records")
        self.assertFalse(xfer[0]['deleted'], "Deleted flag is set")

        db.transfer_destroy(self.ctxt, xfer_id1)
        xfer = db.transfer_get_all(context.get_admin_context())
        self.assertEqual(1, len(xfer), "Unexpected number of transfer records")
        self.assertEqual(xfer[0]['id'], xfer_id2,
                         "Unexpected value for Transfer id")

        nctxt = context.RequestContext(user_id=fake.USER2_ID,
                                       project_id=fake.PROJECT2_ID)
        self.assertRaises(exception.TransferNotFound,
                          db.transfer_destroy, nctxt, xfer_id2)

        db.transfer_destroy(nctxt.elevated(), xfer_id2)
        xfer = db.transfer_get_all(context.get_admin_context())
        self.assertEqual(0, len(xfer), "Unexpected number of transfer records")

    def test_transfer_accept(self):
        volume = utils.create_volume(self.ctxt)
        xfer_id = self._create_transfer(volume['id'], volume['project_id'])
        nctxt = context.RequestContext(user_id=fake.USER2_ID,
                                       project_id=fake.PROJECT2_ID)
        xfer = db.transfer_get(nctxt.elevated(), xfer_id)
        self.assertEqual(volume.project_id, xfer['source_project_id'])
        self.assertFalse(xfer['accepted'])
        self.assertIsNone(xfer['destination_project_id'])
        db.transfer_accept(nctxt.elevated(), xfer_id, fake.USER2_ID,
                           fake.PROJECT2_ID)

        nctxt_admin = nctxt.elevated()

        with db_api.main_context_manager.reader.using(nctxt_admin):
            xfer = db_api.model_query(
                nctxt_admin, models.Transfer, read_deleted='yes'
            ).filter_by(id=xfer_id).first()

        self.assertEqual(volume.project_id, xfer['source_project_id'])
        self.assertTrue(xfer['accepted'])
        self.assertEqual(fake.PROJECT2_ID, xfer['destination_project_id'])

    def test_transfer_accept_with_snapshots(self):
        volume_id = utils.create_volume(self.ctxt)['id']
        snapshot_id1 = utils.create_snapshot(self.ctxt, volume_id,
                                             status='available')['id']
        snapshot_id2 = utils.create_snapshot(self.ctxt, volume_id,
                                             status='available')['id']
        xfer_id = self._create_transfer(volume_id)
        nctxt = context.RequestContext(user_id=fake.USER2_ID,
                                       project_id=fake.PROJECT2_ID)
        db.transfer_accept(nctxt.elevated(), xfer_id, fake.USER2_ID,
                           fake.PROJECT2_ID)
        self.assertEqual(fake.PROJECT2_ID,
                         db.snapshot_get(nctxt, snapshot_id1)['project_id'])
        self.assertEqual(fake.PROJECT2_ID,
                         db.snapshot_get(nctxt, snapshot_id2)['project_id'])

    def test_transfer_accept_with_snapshots_invalid_status(self):
        volume_id = utils.create_volume(self.ctxt)['id']
        snapshot_id1 = utils.create_snapshot(self.ctxt, volume_id,
                                             status='available')['id']
        snapshot_id2 = utils.create_snapshot(self.ctxt, volume_id)['id']
        xfer_id = self._create_transfer(volume_id)
        nctxt = context.RequestContext(user_id=fake.USER2_ID,
                                       project_id=fake.PROJECT2_ID)
        self.assertRaises(exception.InvalidSnapshot, db.transfer_accept,
                          nctxt.elevated(), xfer_id, fake.USER2_ID,
                          fake.PROJECT2_ID)
        self.assertEqual(fake.PROJECT_ID,
                         db.snapshot_get(self.ctxt,
                                         snapshot_id1)['project_id'])
        self.assertEqual(fake.PROJECT_ID,
                         db.snapshot_get(self.ctxt,
                                         snapshot_id2)['project_id'])
        self.assertEqual('awaiting-transfer',
                         db.volume_get(self.ctxt, volume_id)['status'])



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from cinder.exception import VolumeBackendAPIException


class TatlinAPIException(VolumeBackendAPIException):
    path = ''
    code = 500
    message = ''

    def __init__(self, code, message, path=''):
        self.code = code
        self.message = message
        self.path = path
